from typing import Any

import boto3
from botocore.exceptions import ClientError

from dbt_platform_helper.platform_exception import PlatformException


class AutoscalingProvider:
    def __init__(self, client: boto3.client):
        self.autoscaling_client = client

    def describe_autoscaling_target(
        self, cluster_name: str, ecs_service_name: str
    ) -> dict[str, Any]:
        """Return autoscaling target information for an ECS service."""

        try:
            response = self.autoscaling_client.describe_scalable_targets(
                ServiceNamespace="ecs", ResourceIds=[f"service/{cluster_name}/{ecs_service_name}"]
            )
            return response["ScalableTargets"][0]
        except ClientError as err:
            raise PlatformException(f"Error retrieving scalable targets: {err}")



from schema import SchemaError

from dbt_platform_helper.entities.platform_config_schema import PlatformConfigSchema
from dbt_platform_helper.providers.config_validator import ConfigValidator


def validate_addons(addons: dict):
    """
    Validate the addons file and return a dictionary of addon: error message.
    """
    errors = {}

    for addon_name, addon in addons.items():
        try:
            addon_type = addon.get("type", None)
            if not addon_type:
                errors[addon_name] = f"Missing addon type in addon '{addon_name}'"
                continue
            schema = PlatformConfigSchema.extension_schemas().get(addon_type, None)
            if not schema:
                errors[addon_name] = (
                    f"Unsupported addon type '{addon_type}' in addon '{addon_name}'"
                )
                continue
            schema.validate(addon)
        except SchemaError as ex:
            errors[addon_name] = f"Error in {addon_name}: {ex.code}"

    ConfigValidator().validate_supported_redis_versions({"extensions": addons})
    ConfigValidator().validate_supported_opensearch_versions({"extensions": addons})

    return errors



from typing import Callable

import boto3

from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.aws.opensearch import Opensearch
from dbt_platform_helper.providers.aws.redis import Redis
from dbt_platform_helper.providers.cache import Cache
from dbt_platform_helper.providers.cache import GetAWSVersionStrategy
from dbt_platform_helper.providers.io import ClickIOProvider


class ConfigValidatorError(PlatformException):
    pass


class ConfigValidator:

    def __init__(
        self,
        validations: Callable[[dict], None] = None,
        io: ClickIOProvider = ClickIOProvider(),
        session: boto3.Session = None,
    ):
        self.validations = validations or [
            self.validate_supported_redis_versions,
            self.validate_supported_opensearch_versions,
            self.validate_environment_pipelines,
            self.validate_environment_pipelines_triggers,
            self.validate_database_copy_section,
            self.validate_s3_data_migration_config,
            self.validate_cache_invalidation_config,
        ]
        self.io = io
        self.session = session

    def run_validations(self, config: dict):
        for validation in self.validations:
            validation(config)

    def _validate_extension_supported_versions(
        self, config, aws_provider, extension_type, version_key
    ):
        extensions = config.get("extensions", {})
        if not extensions:
            return

        extensions_for_type = [
            extension
            for extension in config.get("extensions", {}).values()
            if extension.get("type") == extension_type
        ]

        # In this format so it can be monkey patched initially via mock_get_data fixture
        cache_provider = Cache()
        get_data_strategy = GetAWSVersionStrategy(aws_provider)
        supported_extension_versions = cache_provider.get_data(get_data_strategy)
        extensions_with_invalid_version = []

        for extension in extensions_for_type:

            environments = extension.get("environments", {})

            if not isinstance(environments, dict):
                self.io.error(
                    f"{extension_type} extension definition is invalid type, expected dictionary",
                )
                continue
            for environment, env_config in environments.items():

                # An extension version doesn't need to be specified for all environments, provided one is specified under "*".
                # So check if the version is set before checking if it's supported
                extension_version = env_config.get(version_key)
                if extension_version and extension_version not in supported_extension_versions:
                    extensions_with_invalid_version.append(
                        {"environment": environment, "version": extension_version}
                    )

        for version_failure in extensions_with_invalid_version:
            self.io.error(
                f"{extension_type} version for environment {version_failure['environment']} is not in the list of supported {extension_type} versions: {supported_extension_versions}. Provided Version: {version_failure['version']}",
            )

    def _get_client(self, service_name: str):
        if self.session:
            return self.session.client(service_name)
        return boto3.client(service_name)

    def validate_supported_redis_versions(self, config):
        return self._validate_extension_supported_versions(
            config=config,
            aws_provider=Redis(self._get_client("elasticache")),
            extension_type="redis",  # TODO: DBTP-1888: this is information which can live in the RedisProvider
            version_key="engine",  # TODO: DBTP-1888: this is information which can live in the RedisProvider
        )

    def validate_supported_opensearch_versions(self, config):
        return self._validate_extension_supported_versions(
            config=config,
            aws_provider=Opensearch(self._get_client("opensearch")),
            extension_type="opensearch",  # TODO: DBTP-1888: this is information which can live in the OpensearchProvider
            version_key="engine",  # TODO: DBTP-1888: this is information which can live in the OpensearchProvider
        )

    def validate_environment_pipelines(self, config):
        bad_pipelines = {}
        for pipeline_name, pipeline in config.get("environment_pipelines", {}).items():
            bad_envs = []
            pipeline_account = pipeline.get("account", None)
            if pipeline_account:
                for env in pipeline.get("environments", {}).keys():
                    env_account = (
                        config.get("environments", {})
                        .get(env, {})
                        .get("accounts", {})
                        .get("deploy", {})
                        .get("name")
                    )
                    if not env_account == pipeline_account:
                        bad_envs.append(env)
            if bad_envs:
                bad_pipelines[pipeline_name] = {"account": pipeline_account, "bad_envs": bad_envs}
        if bad_pipelines:
            message = "The following pipelines are misconfigured:"
            for pipeline, detail in bad_pipelines.items():
                envs = detail["bad_envs"]
                acc = detail["account"]
                message += f"  '{pipeline}' - these environments are not in the '{acc}' account: {', '.join(envs)}\n"
            raise ConfigValidatorError(message)

    def validate_environment_pipelines_triggers(self, config):
        errors = []
        pipelines_with_triggers = {
            pipeline_name: pipeline
            for pipeline_name, pipeline in config.get("environment_pipelines", {}).items()
            if "pipeline_to_trigger" in pipeline
        }

        for pipeline_name, pipeline in pipelines_with_triggers.items():
            pipeline_to_trigger = pipeline["pipeline_to_trigger"]
            if pipeline_to_trigger not in config.get("environment_pipelines", {}):
                message = f"  '{pipeline_name}' - '{pipeline_to_trigger}' is not a valid target pipeline to trigger"

                errors.append(message)
                continue

            if pipeline_to_trigger == pipeline_name:
                message = f"  '{pipeline_name}' - pipelines cannot trigger themselves"
                errors.append(message)

        if errors:
            error_message = "The following pipelines are misconfigured: \n"
            raise ConfigValidatorError(error_message + "\n  ".join(errors))

    def validate_database_copy_section(self, config):
        extensions = config.get("extensions", {})
        if not extensions:
            return

        postgres_extensions = {
            key: ext for key, ext in extensions.items() if ext.get("type", None) == "postgres"
        }

        if not postgres_extensions:
            return

        errors = []

        for extension_name, extension in postgres_extensions.items():
            database_copy_sections = extension.get("database_copy", [])

            if not database_copy_sections:
                return

            all_environments = [
                env for env in config.get("environments", {}).keys() if not env == "*"
            ]
            all_envs_string = ", ".join(all_environments)

            for section in database_copy_sections:
                from_env = section["from"]
                to_env = section["to"]

                if from_env == to_env:
                    errors.append(
                        f"database_copy 'to' and 'from' cannot be the same environment in extension '{extension_name}'."
                    )

                if "prod" in to_env:
                    errors.append(
                        f"Copying to a prod environment is not supported: database_copy 'to' cannot be '{to_env}' in extension '{extension_name}'."
                    )

                if from_env not in all_environments:
                    errors.append(
                        f"database_copy 'from' parameter must be a valid environment ({all_envs_string}) but was '{from_env}' in extension '{extension_name}'."
                    )

                if to_env not in all_environments:
                    errors.append(
                        f"database_copy 'to' parameter must be a valid environment ({all_envs_string}) but was '{to_env}' in extension '{extension_name}'."
                    )

        if errors:
            raise ConfigValidatorError("\n".join(errors))

    def validate_s3_data_migration_config(self, config: dict):
        extensions = config.get("extensions", {})
        if not extensions:
            return

        s3_extensions = {
            key: ext for key, ext in extensions.items() if ext.get("type", None) == "s3"
        }

        if not s3_extensions:
            return

        errors = []

        for extension_name, extension in s3_extensions.items():
            for env, env_config in extension.get("environments", {}).items():
                if "data_migration" not in env_config:
                    continue
                data_migration = env_config.get("data_migration", {})
                if extension.get("serve_static_content", {}):
                    errors.append(
                        "Data migration is not supported for static S3 buckets to avoid the risk of unintentionally exposing private data. However, you can copy data on an ad hoc basis using AWS CLI commands such as 'aws s3 sync' or 'aws s3 cp'."
                    )
                if "import" in data_migration and "import_sources" in data_migration:
                    errors.append(
                        f"Error in '{extension_name}.environments.{env}.data_migration': only the 'import_sources' property is required - 'import' is deprecated."
                    )
                if "import" not in data_migration and "import_sources" not in data_migration:
                    errors.append(
                        f"'import_sources' property in '{extension_name}.environments.{env}.data_migration' is missing."
                    )
        if errors:
            raise ConfigValidatorError("\n".join(errors))

    def validate_cache_invalidation_config(self, config: dict):
        codebase_pipelines = config.get("codebase_pipelines")
        if not codebase_pipelines:
            return

        errors = []

        all_environments = [env for env in config.get("environments", {}).keys() if not env == "*"]

        for codebase in codebase_pipelines.values():
            cache_invalidation_config = codebase.get("cache_invalidation")
            if cache_invalidation_config:
                for domain, config in cache_invalidation_config.get("domains").items():
                    environment = config.get("environment")
                    if environment not in all_environments:
                        errors.append(
                            f"Error in cache invalidation configuration for the domain '{domain}'.  Environment '{environment}' is not defined for this application"
                        )

        if errors:
            raise ConfigValidatorError("\n".join(errors))



import click

from dbt_platform_helper.domain.config import Config
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.aws.sso_auth import SSOAuthProvider
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.utils.aws import get_aws_session_or_abort
from dbt_platform_helper.utils.click import ClickDocOptGroup


@click.group(cls=ClickDocOptGroup)
def config():
    """Perform actions on configuration files."""


@config.command()
def validate():
    """Validate deployment or application configuration."""
    try:
        Config().validate()
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))


@config.command()
def migrate():
    """Update configuration to match current schema."""
    try:
        Config().migrate()
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))


@config.command()
@click.option("--file-path", "-fp", default="~/.aws/config")
def aws(file_path):
    """
    Writes a local config file containing all the AWS profiles to which the
    logged in user has access.

    If no `--file-path` is specified, defaults to `~/.aws/config`.
    """
    try:
        session = get_aws_session_or_abort()
        Config(sso=SSOAuthProvider(session)).generate_aws(file_path)
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))



from dbt_platform_helper.platform_exception import PlatformException


class AWSException(PlatformException):
    pass


class CreateTaskTimeoutException(AWSException):
    def __init__(self, addon_name: str, application_name: str, environment: str):
        super().__init__(
            f"""Client ({addon_name}) ECS task has failed to start for "{application_name}" in "{environment}" environment."""
        )


IMAGE_NOT_FOUND_TEMPLATE = """An image labelled "{image_ref}" could not be found in your image repository. Try the `platform-helper codebase build` command first."""


class ImageNotFoundException(AWSException):
    def __init__(self, image_ref: str):
        super().__init__(IMAGE_NOT_FOUND_TEMPLATE.format(image_ref=image_ref))


MULTIPLE_IMAGES_FOUND_TEMPLATE = (
    'Image reference "{image_ref}" is matched by the following images: {matching_images}'
)


class MultipleImagesFoundException(AWSException):
    def __init__(self, image_ref: str, matching_images: list[str]):
        super().__init__(
            MULTIPLE_IMAGES_FOUND_TEMPLATE.format(
                image_ref=image_ref, matching_images=", ".join(sorted(matching_images))
            )
        )


REPOSITORY_NOT_FOUND_TEMPLATE = """The ECR repository "{repository}" could not be found."""


class RepositoryNotFoundException(AWSException):
    def __init__(self, repository: str):
        super().__init__(REPOSITORY_NOT_FOUND_TEMPLATE.format(repository=repository))


class LogGroupNotFoundException(AWSException):
    def __init__(self, log_group_name: str):
        super().__init__(f"""No log group called "{log_group_name}".""")


# TODO: DBTP-1976: This should probably be in the AWS Copilot provider, but was causing circular import when we tried it pre refactoring the utils/aws.py
class CopilotCodebaseNotFoundException(PlatformException):
    def __init__(self, codebase: str):
        super().__init__(
            f"""The codebase "{codebase}" either does not exist or has not been deployed."""
        )


class CreateAccessTokenException(AWSException):
    def __init__(self, client_id: str):
        super().__init__(f"""Failed to create access token for Client "{client_id}".""")


class UnableToRetrieveSSOAccountList(AWSException):
    def __init__(self):
        super().__init__("Unable to retrieve AWS SSO account list")


class UnableToRetrieveSSOAccountRolesList(AWSException):
    def __init__(self, account_id: str):
        super().__init__(f"Unable to retrieve AWS SSO roles list for AWS account {account_id}")



import copy
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from typing import List

from dbt_platform_helper.constants import COPILOT_RULE_PRIORITY
from dbt_platform_helper.constants import DUMMY_RULE_REASON
from dbt_platform_helper.constants import MAINTENANCE_PAGE_REASON
from dbt_platform_helper.constants import MANAGED_BY_PLATFORM
from dbt_platform_helper.constants import MANAGED_BY_SERVICE_TERRAFORM
from dbt_platform_helper.constants import PLATFORM_RULE_STARTING_PRIORITY
from dbt_platform_helper.constants import RULE_PRIORITY_INCREMENT
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.config_validator import ConfigValidator
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.load_balancers import LoadBalancerProvider
from dbt_platform_helper.utils.application import load_application


class RollbackException(PlatformException):
    pass


class RuleType(Enum):
    PLATFORM = "platform"
    MAINTENANCE = "maintenance"
    DEFAULT = "default"
    COPILOT = "copilot"
    MANUAL = "manual"
    DUMMY = "dummy"


class Deployment(Enum):
    PLATFORM = "platform"
    COPILOT = "copilot"
    DUAL_DEPLOY_PLATFORM = "dual-deploy-platform-traffic"
    DUAL_DEPLOY_COPILOT = "dual-deploy-copilot-traffic"


@dataclass
class OperationState:
    created_rules: List[str] = field(default_factory=list)
    deleted_rules: List[object] = field(default_factory=list)
    listener_arn: str = ""


class UpdateALBRules:

    def __init__(
        self,
        session,
        config_provider: ConfigProvider = ConfigProvider(ConfigValidator()),
        io: ClickIOProvider = ClickIOProvider(),
        load_application=load_application,
        load_balancer_p: LoadBalancerProvider = LoadBalancerProvider,
    ):
        self.config_provider = config_provider
        self.io = io
        self.load_application = load_application
        self.load_balancer: LoadBalancerProvider = load_balancer_p(session, io=self.io)

    def update_alb_rules(
        self,
        environment: str,
    ):
        """
        Change ALB rules for a given environment.

        Attempt to rollback the rules created/deleted if a failure occurs.
        """

        operation_state = OperationState()

        try:
            self._execute_rule_updates(environment, operation_state)
            if operation_state.created_rules:
                self.io.info(f"Created rules: {operation_state.created_rules}")
            if operation_state.deleted_rules:
                deleted_arns = [rule["RuleArn"] for rule in operation_state.deleted_rules]
                self.io.info(f"Deleted rules: {deleted_arns}")
        except Exception as e:
            if operation_state.created_rules or operation_state.deleted_rules:
                self.io.error(f"Error during rule update: {str(e)}")
                self.io.warn("Rolling back")
                self.io.info("Attempting to rollback changes ...")
                try:
                    self._rollback_changes(operation_state)
                except RollbackException as rollback_error:
                    raise PlatformException(f"Rollback failed: \n{str(rollback_error)}")
            else:
                raise

    def _execute_rule_updates(self, environment: str, operation_state: OperationState):
        platform_config = self.config_provider.get_enriched_config()

        application_name = platform_config.get("application", "")
        application = self.load_application(app=application_name)

        service_deployment_mode = (
            platform_config.get("environments")
            .get(environment, {})
            .get("service-deployment-mode", Deployment.COPILOT.value)
        )

        self.io.info(f"Deployment Mode: {service_deployment_mode}")

        listener_arn = self.load_balancer.get_https_listener_for_application(
            application.name, environment
        )

        operation_state.listener_arn = listener_arn

        self.io.debug(f"Listener ARN: {listener_arn}")

        rules = self.load_balancer.get_rules_with_tags_by_listener_arn(listener_arn)

        mapped_rules = {
            rule_type: [rule for rule in rules if self._filter_rule_type(rule) == rule_type]
            for rule_type in set(self._filter_rule_type(rule) for rule in rules)
        }

        if mapped_rules.get(RuleType.MANUAL.value, []):
            rule_arns = [rule["RuleArn"] for rule in mapped_rules[RuleType.MANUAL.value]]
            message = f"""The following rules have been created manually please review and if required set 
            the rules priority to the copilot range after priority: {COPILOT_RULE_PRIORITY}.\n
            Rules: {rule_arns}"""
            raise PlatformException(message)

        if (
            service_deployment_mode == Deployment.PLATFORM.value
            or service_deployment_mode == Deployment.DUAL_DEPLOY_PLATFORM.value
        ):

            if len(mapped_rules.get(RuleType.PLATFORM.value, [])) > 0 and len(
                mapped_rules.get(RuleType.PLATFORM.value, [])
            ) != len(mapped_rules.get(RuleType.COPILOT.value, [])):
                raise PlatformException("Platform rules are partially created, please review.")

            if len(mapped_rules.get(RuleType.PLATFORM.value, [])) == len(
                mapped_rules.get(RuleType.COPILOT.value, [])
            ):
                self.io.info("Platform rules already exist, skipping creation")
                return  # early exit

            grouped = dict()

            service_mapped_tgs = self._get_tg_arns_for_platform_services(
                application_name, environment
            )
            for copilot_rule in mapped_rules.get(RuleType.COPILOT.value, []):
                rule_arn = copilot_rule["RuleArn"]
                self.io.debug(f"Building platform rule for corresponding copilot rule: {rule_arn}")
                sorted_hosts = sorted(copilot_rule["Conditions"].get("host-header", []))
                # Depth represents the specificity of the path condition, allowing us to sort in decreasing complexity.
                path_depth = max(
                    [
                        len([sub_path for sub_path in path.split("/") if sub_path])
                        for path in copilot_rule["Conditions"].get("path-pattern", [])
                    ]
                )
                list_conditions = [
                    {"Field": key, "Values": value}
                    for key, value in copilot_rule["Conditions"].items()
                ]

                service_name = self._get_service_from_tg(copilot_rule)

                tg_arn = service_mapped_tgs[service_name]

                actions = self._create_new_actions(copilot_rule["Actions"], tg_arn)
                self.io.debug(f"Updated forward action for service {service_name} to use: {tg_arn}")
                if grouped.get(",".join(sorted_hosts)):
                    grouped[",".join(sorted_hosts)].append(
                        {
                            "copilot_rule": rule_arn,
                            "actions": actions,
                            "conditions": list_conditions,
                            "path_depth": path_depth,
                            "service_name": service_name,
                        }
                    )
                else:
                    grouped[",".join(sorted_hosts)] = [
                        {
                            "copilot_rule": rule_arn,
                            "actions": actions,
                            "conditions": list_conditions,
                            "path_depth": path_depth,
                            "service_name": service_name,
                        }
                    ]

            rule_priority = PLATFORM_RULE_STARTING_PRIORITY
            for hosts, rules in grouped.items():
                rules.sort(key=lambda x: x["path_depth"], reverse=True)

                for rule in rules:
                    # Create rule with priority
                    copilot_rule = rule.get("copilot_rule", "")
                    self.io.debug(
                        f"Creating platform rule for corresponding copilot rule: {copilot_rule}"
                    )
                    rule_arn = self.load_balancer.create_rule(
                        listener_arn,
                        rule["actions"],
                        rule["conditions"],
                        rule_priority,
                        tags=[
                            {"Key": "application", "Value": application_name},
                            {"Key": "environment", "Value": environment},
                            {"Key": "service", "Value": rule["service_name"]},
                            {"Key": "reason", "Value": "service"},
                            {"Key": "managed-by", "Value": MANAGED_BY_PLATFORM},
                        ],
                    )["Rules"][0]["RuleArn"]
                    operation_state.created_rules.append(rule_arn)
                    rule_priority += RULE_PRIORITY_INCREMENT

                next_thousandth = lambda x: ((x // 1000) + 1) * 1000
                rule_priority = next_thousandth(rule_priority)

        if (
            service_deployment_mode == Deployment.COPILOT.value
            or service_deployment_mode == Deployment.DUAL_DEPLOY_COPILOT.value
        ):
            self._delete_rules(mapped_rules.get(RuleType.PLATFORM.value, []), operation_state)

    def _delete_rules(self, rules: List[dict], operation_state: OperationState):
        for rule in rules:
            rule_arn = rule["RuleArn"]
            try:
                self.load_balancer.delete_listener_rule_by_resource_arn(rule_arn)
                operation_state.deleted_rules.append(rule)
                self.io.debug(f"Deleted existing rule: {rule_arn}")
            except Exception as e:
                self.io.error(f"Failed to delete existing rule {rule_arn}: {str(e)}")
                raise

    def _filter_rule_type(self, rule: dict) -> str:
        if rule["Tags"]:
            if rule["Tags"].get("managed-by", "") == MANAGED_BY_PLATFORM:
                return RuleType.PLATFORM.value
            if rule["Tags"].get("reason", None) == MAINTENANCE_PAGE_REASON:
                return RuleType.MAINTENANCE.value
            if rule["Tags"].get("reason", None) == DUMMY_RULE_REASON:
                return RuleType.DUMMY.value

        if rule["Priority"] == "default":
            return RuleType.DEFAULT.value
        if int(rule["Priority"]) >= COPILOT_RULE_PRIORITY:
            return RuleType.COPILOT.value

        return RuleType.MANUAL.value

    def _get_service_from_tg(self, rule: dict) -> str:
        target_group_arn = None

        for action in rule["Actions"]:
            if action["Type"] == "forward":
                target_group_arn = action["TargetGroupArn"]

        if target_group_arn:
            try:
                tgs = self.load_balancer.get_target_groups_with_tags([target_group_arn])
                return tgs[0]["Tags"].get("copilot-service", None)
            except IndexError:
                raise PlatformException(f"No target group found for arn: {target_group_arn}")
        else:
            rule_arn = rule["RuleArn"]
            raise PlatformException(f"No target group arn found in rule: {rule_arn}")

    def _get_tg_arns_for_platform_services(
        self, application: str, environment: str, managed_by: str = MANAGED_BY_SERVICE_TERRAFORM
    ) -> dict:
        tgs = self.load_balancer.get_target_groups_with_tags([])
        service_mapped_tgs = {}
        for tg in tgs:
            if (
                tg["Tags"].get("application", "") == application
                and tg["Tags"].get("environment", "") == environment
                and tg["Tags"].get("managed-by", "") == managed_by
            ):
                if tg["Tags"].get("service", ""):
                    service_mapped_tgs[tg["Tags"].get("service")] = tg["TargetGroupArn"]
                else:
                    tg_name = tg["TargetGroupName"]
                    self.io.warn(f"Target group {tg_name} has no 'service' tag")
        return service_mapped_tgs

    def _create_new_actions(self, actions: dict, tg_arn: str) -> dict:

        updated_actions = copy.deepcopy(actions)
        for action in updated_actions:
            if action.get("Type") == "forward" and "TargetGroupArn" in action:
                action["TargetGroupArn"] = tg_arn
                for tg in action["ForwardConfig"]["TargetGroups"]:
                    tg["TargetGroupArn"] = tg_arn
        return updated_actions

    def _rollback_changes(self, operation_state: OperationState) -> bool:
        rollback_errors = []
        delete_rollbacks = []
        create_rollbacks = []
        for rule_arn in operation_state.created_rules:
            try:
                self.io.debug(f"Rolling back: Deleting created rule {rule_arn}")
                self.load_balancer.delete_listener_rule_by_resource_arn(rule_arn)
                delete_rollbacks.append(rule_arn)
            except Exception as e:
                error_msg = f"Failed to delete rule {rule_arn} during rollback: {str(e)}"
                rollback_errors.append(error_msg)

        for rule_snapshot in operation_state.deleted_rules:
            rule_arn = rule_snapshot["RuleArn"]
            try:
                self.io.debug(f"Rolling back: Recreating deleted rule {rule_arn}")
                create_rollbacks.append(
                    self.load_balancer.create_rule(
                        operation_state.listener_arn,
                        actions=rule_snapshot["Actions"],
                        conditions=[
                            {"Field": key, "Values": value}
                            for key, value in rule_snapshot["Conditions"].items()
                        ],
                        priority=int(rule_snapshot["Priority"]),
                        tags=[
                            {"Key": key, "Value": value}
                            for key, value in rule_snapshot["Tags"].items()
                        ],
                    )["Rules"][0]["RuleArn"]
                )
            except Exception as e:
                error_msg = f"Failed to recreate rule {rule_arn} during rollback: {str(e)}"
                rollback_errors.append(error_msg)

        if rollback_errors:
            self.io.warn("Some rollback operations failed. Manual intervention may be required.")
            errors = "\n".join(rollback_errors)
            raise RollbackException(f"Rollback partially failed: {errors}")
        else:
            self.io.info("Rollback completed successfully")
            self.io.info(
                f"Rolledback rules by creating: {create_rollbacks} \n and deleting {delete_rollbacks}"
            )



from dataclasses import dataclass

from dbt_platform_helper.entities.semantic_version import SemanticVersion
from dbt_platform_helper.platform_exception import PlatformException


class UnsupportedVersionException(PlatformException):
    def __init__(self, version: str):
        super().__init__(
            f"""Platform-helper version {version} is not compatible with platform-helper. Please install version platform-helper version 14 or later."""
        )


@dataclass
class VersionStatus:
    installed: SemanticVersion = None
    latest: SemanticVersion = None

    def __str__(self):
        attrs = {
            key: value for key, value in vars(self).items() if isinstance(value, SemanticVersion)
        }
        attrs_str = ", ".join(f"{key}: {value}" for key, value in attrs.items())
        return f"{self.__class__.__name__}: {attrs_str}"

    def is_outdated(self):
        return self.installed != self.latest



import json
import os
import time
import urllib.parse
from configparser import ConfigParser
from pathlib import Path

import boto3
import botocore
import botocore.exceptions
import click
from boto3 import Session
from botocore.exceptions import ClientError

from dbt_platform_helper.constants import REFRESH_TOKEN_MESSAGE
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.aws.exceptions import (
    CopilotCodebaseNotFoundException,
)
from dbt_platform_helper.providers.aws.exceptions import LogGroupNotFoundException
from dbt_platform_helper.providers.validation import ValidationException

SSM_BASE_PATH = "/copilot/{app}/{env}/secrets/"
SSM_PATH = "/copilot/{app}/{env}/secrets/{name}"
AWS_SESSION_CACHE = {}


def get_aws_session_or_abort(aws_profile: str = None) -> boto3.session.Session:
    aws_profile = aws_profile or os.getenv("AWS_PROFILE")
    if aws_profile in AWS_SESSION_CACHE:
        return AWS_SESSION_CACHE[aws_profile]

    click.secho(f'Checking AWS connection for profile "{aws_profile}"...', fg="cyan")

    try:
        session = boto3.session.Session(profile_name=aws_profile)
        sts = session.client("sts")
        account_id, user_id = get_account_details(sts)
        click.secho("Credentials are valid.", fg="green")

    except botocore.exceptions.ProfileNotFound:
        _handle_error(f'AWS profile "{aws_profile}" is not configured.')
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "ExpiredToken":
            _handle_error(
                f"Credentials are NOT valid.  \nPlease login with: aws sso login --profile {aws_profile}"
            )
    except botocore.exceptions.NoCredentialsError:
        _handle_error("There are no credentials set for this session.", REFRESH_TOKEN_MESSAGE)
    except botocore.exceptions.UnauthorizedSSOTokenError:
        _handle_error("The SSO Token used for this session is unauthorised.", REFRESH_TOKEN_MESSAGE)
    except botocore.exceptions.TokenRetrievalError:
        _handle_error("Unable to retrieve the Token for this session.", REFRESH_TOKEN_MESSAGE)
    except botocore.exceptions.SSOTokenLoadError:
        _handle_error(
            "The SSO session associated with this profile has expired, is not set or is otherwise invalid.",
            REFRESH_TOKEN_MESSAGE,
        )

    alias_client = session.client("iam")
    account_name = alias_client.list_account_aliases().get("AccountAliases", [])

    _log_account_info(account_name, account_id)

    click.echo(
        click.style("User: ", fg="yellow")
        + click.style(f"{user_id.split(':')[-1]}\n", fg="white", bold=True)
    )

    AWS_SESSION_CACHE[aws_profile] = session
    return session


def _handle_error(message: str, refresh_token_message: str = None) -> None:
    full_message = message + (" " + refresh_token_message if refresh_token_message else "")
    click.secho(full_message, fg="red")
    exit(1)


def _log_account_info(account_name: list, account_id: str) -> None:
    if account_name:
        click.echo(
            click.style("Logged in with AWS account: ", fg="yellow")
            + click.style(f"{account_name[0]}/{account_id}", fg="white", bold=True)
        )
    else:
        click.echo(
            click.style("Logged in with AWS account id: ", fg="yellow")
            + click.style(f"{account_id}", fg="white", bold=True)
        )


class NoProfileForAccountIdException(PlatformException):
    def __init__(self, account_id):
        super().__init__(f"No profile found for account {account_id}")


def get_profile_name_from_account_id(account_id: str):
    aws_config = ConfigParser()
    aws_config.read(Path.home().joinpath(".aws/config"))
    for section in aws_config.sections():
        found_account_id = aws_config[section].get(
            "sso_account_id", aws_config[section].get("profile_account_id", None)
        )
        if account_id == found_account_id:
            return section.removeprefix("profile ")

    raise NoProfileForAccountIdException(account_id)


def get_ssm_secret_names(app, env):
    session = get_aws_session_or_abort()
    client = session.client("ssm")

    path = SSM_BASE_PATH.format(app=app, env=env)

    params = dict(
        Path=path,
        Recursive=False,
        WithDecryption=True,
        MaxResults=10,
    )

    secret_names = []

    while True:
        response = client.get_parameters_by_path(**params)

        for secret in response["Parameters"]:
            secret_names.append(secret["Name"])

        if "NextToken" in response:
            params["NextToken"] = response["NextToken"]
        else:
            break

    return sorted(secret_names)


def get_ssm_secrets(app, env, session=None, path=None):
    """Return secrets from AWS Parameter Store as a list of tuples with the
    secret name and secret value."""

    if not session:
        session = get_aws_session_or_abort()
    client = session.client("ssm")

    if not path:
        path = SSM_BASE_PATH.format(app=app, env=env)

    params = dict(
        Path=path,
        Recursive=False,
        WithDecryption=True,
        MaxResults=10,
    )

    secrets = []

    while True:
        try:
            response = client.get_parameters_by_path(**params)
        except ClientError as e:
            if e.response["Error"]["Code"] == "AccessDeniedException":
                click.secho(
                    "Access denied on SSM, due to missing permissions. Please update your environment infrastructure.",
                    fg="magenta",
                )
                break
            else:
                raise e

        for secret in response["Parameters"]:
            secrets.append((secret["Name"], secret["Value"]))

        if "NextToken" in response:
            params["NextToken"] = response["NextToken"]
        else:
            break

    return sorted(secrets)


def set_ssm_param(
    app, env, param_name, param_value, overwrite, exists, description="Copied from Cloud Foundry."
):
    session = get_aws_session_or_abort()
    client = session.client("ssm")

    parameter_args = dict(
        Name=param_name,
        Description=description,
        Value=param_value,
        Type="SecureString",
        Overwrite=overwrite,
        Tags=[
            {"Key": "copilot-application", "Value": app},
            {"Key": "copilot-environment", "Value": env},
        ],
        Tier="Intelligent-Tiering",
    )

    if overwrite and not exists:
        raise ValidationException(
            """Arguments "overwrite" is set to True, but "exists" is set to False."""
        )

    if overwrite and exists:
        # Tags can't be updated when overwriting
        del parameter_args["Tags"]

    client.put_parameter(**parameter_args)


def get_codestar_connection_arn(app_name):
    session = get_aws_session_or_abort()
    response = session.client("codestar-connections").list_connections()

    for connection in response["Connections"]:
        if connection["ConnectionName"] == app_name:
            return connection["ConnectionArn"]


def get_account_details(sts_client=None):
    if not sts_client:
        sts_client = get_aws_session_or_abort().client("sts")
    response = sts_client.get_caller_identity()

    return response["Account"], response["UserId"]


def get_postgres_connection_data_updated_with_master_secret(session, parameter_name, secret_arn):
    # TODO: DBTP-1968: This is pretty much the same as dbt_platform_helper.providers.secrets.Secrets.get_postgres_connection_data_updated_with_master_secret
    ssm_client = session.client("ssm")
    secrets_manager_client = session.client("secretsmanager")
    response = ssm_client.get_parameter(Name=parameter_name, WithDecryption=True)
    parameter_value = response["Parameter"]["Value"]

    parameter_data = json.loads(parameter_value)

    secret_response = secrets_manager_client.get_secret_value(SecretId=secret_arn)
    secret_value = json.loads(secret_response["SecretString"])

    parameter_data["username"] = urllib.parse.quote(secret_value["username"])
    parameter_data["password"] = urllib.parse.quote(secret_value["password"])

    return parameter_data


def get_connection_string(
    session: Session,
    app: str,
    env: str,
    db_identifier: str,
    connection_data=get_postgres_connection_data_updated_with_master_secret,
) -> str:
    addon_name = db_identifier.split(f"{app}-{env}-", 1)[1]
    normalised_addon_name = addon_name.replace("-", "_").upper()
    connection_string_parameter = (
        f"/copilot/{app}/{env}/secrets/{normalised_addon_name}_READ_ONLY_USER"
    )
    master_secret_name = f"/copilot/{app}/{env}/secrets/{normalised_addon_name}_RDS_MASTER_ARN"
    master_secret_arn = session.client("ssm").get_parameter(
        Name=master_secret_name, WithDecryption=True
    )["Parameter"]["Value"]

    conn = connection_data(session, connection_string_parameter, master_secret_arn)

    return f"postgres://{conn['username']}:{conn['password']}@{conn['host']}:{conn['port']}/{conn['dbname']}"


def start_build_extraction(codebuild_client, build_options):
    response = codebuild_client.start_build(**build_options)
    return response["build"]["arn"]


def start_pipeline_and_return_execution_id(codepipeline_client, build_options):
    response = codepipeline_client.start_pipeline_execution(**build_options)
    return response["pipelineExecutionId"]


# TODO: DBTP-1888: This should probably be in the AWS Copilot provider
def check_codebase_exists(session: Session, application, codebase: str):
    try:
        # TODO: DBTP-1968: Can this leverage dbt_platform_helper.providers.secrets.Secrets.get_connection_secret_arn?
        ssm_client = session.client("ssm")
        json.loads(
            ssm_client.get_parameter(
                Name=f"/copilot/applications/{application.name}/codebases/{codebase}"
            )["Parameter"]["Value"]
        )
    except (
        KeyError,
        ValueError,
        ssm_client.exceptions.ParameterNotFound,
        json.JSONDecodeError,
    ):
        raise CopilotCodebaseNotFoundException(codebase)


def get_build_url_from_arn(build_arn: str) -> str:
    _, _, _, region, account_id, project_name, build_id = build_arn.split(":")
    project_name = project_name.removeprefix("build/")
    return (
        f"https://eu-west-2.console.aws.amazon.com/codesuite/codebuild/{account_id}/projects/"
        f"{project_name}/build/{project_name}%3A{build_id}"
    )


def get_build_url_from_pipeline_execution_id(execution_id: str, pipeline_name: str) -> str:

    return f"https://eu-west-2.console.aws.amazon.com/codesuite/codepipeline/pipelines/{pipeline_name}/executions/{execution_id}"


def list_latest_images(ecr_client, ecr_repository_name, codebase_repository, echo):
    paginator = ecr_client.get_paginator("describe_images")
    describe_images_response_iterator = paginator.paginate(
        repositoryName=ecr_repository_name,
        filter={"tagStatus": "TAGGED"},
    )
    images = []
    for page in describe_images_response_iterator:
        images += page["imageDetails"]

    sorted_images = sorted(
        images,
        key=lambda i: i["imagePushedAt"],
        reverse=True,
    )

    MAX_RESULTS = 20

    for image in sorted_images[:MAX_RESULTS]:
        try:
            commit_tag = next(t for t in image["imageTags"] if t.startswith("commit-"))
            if not commit_tag:
                continue

            commit_hash = commit_tag.replace("commit-", "")
            echo(
                f"  - https://github.com/{codebase_repository}/commit/{commit_hash} - published: {image['imagePushedAt']}"
            )
        except StopIteration:
            continue


def wait_for_log_group_to_exist(log_client, log_group_name, attempts=30):
    current_attempts = 0
    log_group_exists = False

    while not log_group_exists and current_attempts < attempts:
        current_attempts += 1

        log_group_response = log_client.describe_log_groups(logGroupNamePrefix=log_group_name)
        log_groups = log_group_response.get("logGroups", [])

        for group in log_groups:
            if group["logGroupName"] == log_group_name:
                log_group_exists = True

        time.sleep(1)

    if not log_group_exists:
        raise LogGroupNotFoundException(log_group_name)


def get_image_build_project(codebuild_client, application, codebase):
    project_name = f"{application}-{codebase}-codebase-image-build"
    response = codebuild_client.batch_get_projects(names=[project_name])

    if bool(response.get("projects")):
        return project_name
    else:
        return f"{application}-{codebase}-codebase-pipeline-image-build"


def get_manual_release_pipeline(codepipeline_client, application, codebase):
    pipeline_name = f"{application}-{codebase}-manual-release"
    try:
        codepipeline_client.get_pipeline(name=pipeline_name)
        return pipeline_name
    except ClientError as e:
        if e.response["Error"]["Code"] == "PipelineNotFoundException":
            return f"{pipeline_name}-pipeline"



# This exception exists so that we can easily catch exceptions
# at the command level where we know we can just output the
# error and abort.
class PlatformException(Exception):
    pass


class ValidationException(PlatformException):
    pass



import time
from typing import Any

import boto3
from botocore.exceptions import ClientError

from dbt_platform_helper.platform_exception import PlatformException


class LogsProvider:

    def __init__(self, client: boto3.client):
        self.client = client

    def check_log_streams_present(self, log_group: str, expected_log_streams: list[str]) -> bool:
        """
        Check whether the logs streams provided exist or not.

        Retry for up to 1 minute.
        """

        found_log_streams = set()
        expected_log_streams = set(expected_log_streams)
        timeout_seconds = 60
        poll_interval_seconds = 5
        deadline_seconds = time.monotonic() + timeout_seconds

        while time.monotonic() < deadline_seconds:

            remaining_log_streams = expected_log_streams - found_log_streams
            if not remaining_log_streams:
                return True

            for log_stream in list(remaining_log_streams):
                try:
                    response = self.client.describe_log_streams(
                        logGroupName=log_group, logStreamNamePrefix=log_stream, limit=1
                    )
                except ClientError as e:
                    code = e.response.get("Error", {}).get("Code")
                    if code == "ResourceNotFoundException":
                        continue  # Log stream not there yet, keep going
                    else:
                        raise PlatformException(
                            f"Failed to check if log stream '{log_stream}' exists due to an error {e}"
                        )

                for ls in response.get("logStreams", []):
                    if ls.get("logStreamName") == log_stream:
                        found_log_streams.add(log_stream)

            if expected_log_streams - found_log_streams:
                time.sleep(poll_interval_seconds)

        missing_log_streams = expected_log_streams - found_log_streams
        raise PlatformException(
            f"Timed out waiting for the following log streams to create: {missing_log_streams}"
        )

    def get_log_stream_events(
        self, log_group: str, log_stream: str, limit: int
    ) -> list[dict[str, Any]]:
        """Return events for a specific log stream."""

        try:
            self.check_log_streams_present(log_group=log_group, expected_log_streams=[log_stream])
            response = self.client.get_log_events(
                logGroupName=log_group, logStreamName=log_stream, limit=limit
            )
            return response["events"]
        except ClientError as err:
            raise PlatformException(f"Error retrieving log stream events: {err}")



import re
from typing import Union

from dbt_platform_helper.providers.validation import ValidationException


class IncompatibleMajorVersionException(ValidationException):
    def __init__(self, app_version: str, check_version: str):
        super().__init__()
        self.app_version = app_version
        self.check_version = check_version


class IncompatibleMinorVersionException(ValidationException):
    def __init__(self, app_version: str, check_version: str):
        super().__init__()
        self.app_version = app_version
        self.check_version = check_version


class SemanticVersion:
    def __init__(self, major: int, minor: int, patch: int):
        self.major = major
        self.minor = minor
        self.patch = patch

    def __str__(self) -> str:
        if self.major is None:
            return "unknown"
        return ".".join([str(s) for s in [self.major, self.minor, self.patch]])

    def __repr__(self) -> str:
        return str(self)

    def __lt__(self, other) -> bool:
        return (self.major, self.minor, self.patch) < (other.major, other.minor, other.patch)

    def __eq__(self, other) -> bool:
        if other is None:
            return False
        return (self.major, self.minor, self.patch) == (other.major, other.minor, other.patch)

    def validate_compatibility_with(self, other):
        if other is None:
            raise ValidationException("Cannot compare NoneType")
        if (self.major == 0 and other.major == 0) and (
            self.minor != other.minor or self.patch != other.patch
        ):
            raise IncompatibleMajorVersionException(str(self), str(other))

        if self.major != other.major:
            raise IncompatibleMajorVersionException(str(self), str(other))

        if self.minor != other.minor:
            raise IncompatibleMinorVersionException(str(self), str(other))

    @staticmethod
    def _cast_to_int_with_fallback(input, fallback=-1):
        try:
            return int(input)
        except ValueError:
            return fallback

    @classmethod
    def from_string(self, version_string: Union[str, None]):
        if version_string is None:
            return None

        version_segments = re.split(r"[.\-]", version_string.replace("v", ""))

        if len(version_segments) != 3:
            return None

        major, minor, patch = [self._cast_to_int_with_fallback(s) for s in version_segments]

        return SemanticVersion(major, minor, patch)

    @staticmethod
    def is_semantic_version(version_string):
        valid_semantic_string_regex = r"(?i)^v?[0-9]+[.-][0-9]+[.-][0-9]+$"
        return re.match(valid_semantic_string_regex, version_string)



from collections import OrderedDict
from pathlib import Path

import yaml
from yaml.parser import ParserError
from yamllint import linter
from yamllint.config import YamlLintConfig


class FileProviderException(Exception):
    pass


class YamlFileProviderException(FileProviderException):
    pass


class FileNotFoundException(FileProviderException):
    pass


class InvalidYamlException(YamlFileProviderException):
    def __init__(self, path: str):
        super().__init__(f"""{path} is not valid YAML.""")


class DuplicateKeysException(YamlFileProviderException):
    def __init__(self, duplicate_keys: str):
        super().__init__(f"""Duplicate keys found in your config file: {duplicate_keys}.""")


class YamlFileProvider:
    @staticmethod
    def load(path: str) -> dict:
        """
        Raises:
            FileNotFoundException: file is not there
            InvalidYamlException: file contains invalid yaml
            DuplicateKeysException: yaml contains duplicate keys
        """
        if not Path(path).exists():
            raise FileNotFoundException(f"`{path}` is missing.")
        try:
            yaml_content = yaml.safe_load(Path(path).read_text())
        except ParserError:
            raise InvalidYamlException(path)

        if not yaml_content:
            return {}
        YamlFileProvider.lint_yaml_for_duplicate_keys(path)

        return yaml_content

    @staticmethod
    def write(path: str, contents: dict, comment: str = ""):
        with open(path, "w") as file:
            file.write(comment)
            yaml.add_representer(str, account_number_representer)
            yaml.add_representer(type(None), null_value_representer)

            yaml.dump(
                contents,
                file,
                canonical=False,
                sort_keys=False,
                default_style=None,
                default_flow_style=False,
            )

    @staticmethod
    def lint_yaml_for_duplicate_keys(path):
        duplicate_keys = []
        with open(path, "r") as yaml_file:
            file_contents = yaml_file.read()
            results = linter.run(
                file_contents, YamlLintConfig(yaml.dump({"rules": {"key-duplicates": "enable"}}))
            )
            duplicate_keys = [
                "\t"
                + f"Line {result.line}: {result.message}".replace(
                    " in mapping (key-duplicates)", ""
                )
                for result in results
            ]
        if duplicate_keys:
            raise DuplicateKeysException(",".join(duplicate_keys))

    @staticmethod
    def remove_empty_keys(config: (dict, OrderedDict)) -> (dict, OrderedDict):
        cleaned = config.__class__()

        for k, v in config.items():
            if isinstance(v, (dict, OrderedDict)):
                v = YamlFileProvider.remove_empty_keys(v)
            if v not in (None, [], {}, ()):
                cleaned[k] = v

        return cleaned

    @staticmethod
    def find_and_replace(config, strings: list, replacements: list):
        if len(strings) != len(replacements):
            raise ValueError("'strings' and 'replacements' must be the same length.")
        if not isinstance(strings, list) or not isinstance(replacements, list):
            raise ValueError("'strings' and 'replacements' must both be lists.")
        if isinstance(config, (dict, OrderedDict)):
            return {
                k: YamlFileProvider.find_and_replace(v, strings, replacements)
                for k, v in config.items()
            }
        elif isinstance(config, list):
            return [
                YamlFileProvider.find_and_replace(item, strings, replacements) for item in config
            ]
        elif isinstance(config, str):
            for s, r in zip(strings, replacements):
                config = config.replace(s, r)
            return config
        else:
            return replacements if config == strings else config


def account_number_representer(dumper, data):
    if data.isdigit():
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="'")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style=None)


def null_value_representer(dumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:null", "")



import boto3


class KMSProvider:
    """A provider class for interacting with the AWS KMS (Key Management
    Service)."""

    def __init__(self, kms_client: boto3.client):
        self.kms_client = kms_client

    def describe_key(self, alias_name: str) -> dict:
        """
        Retrieves metadata about a KMS key using its alias.

        Args:
            alias_name (str): The alias name of the KMS key.

        Returns:
            dict: A dictionary containing metadata about the specified KMS key.
        """
        # The kms client can take an alias name as the KeyId
        return self.kms_client.describe_key(KeyId=alias_name)



from dbt_platform_helper.constants import PLATFORM_HELPER_VERSION_OVERRIDE_KEY
from dbt_platform_helper.constants import (
    TERRAFORM_EXTENSIONS_MODULE_SOURCE_OVERRIDE_ENV_VAR,
)
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.environment_variable import (
    EnvironmentVariableProvider,
)
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.terraform_manifest import TerraformManifestProvider


class TerraformEnvironmentException(PlatformException):
    pass


class EnvironmentNotFoundException(TerraformEnvironmentException):
    pass


class TerraformEnvironment:
    def __init__(
        self,
        config_provider,
        manifest_provider: TerraformManifestProvider = None,
        environment_variable_provider: EnvironmentVariableProvider = None,
        io: ClickIOProvider = ClickIOProvider(),
        platform_helper_version_override: str = None,
    ):
        self.io = io
        self.config_provider = config_provider
        self.manifest_provider = manifest_provider or TerraformManifestProvider()
        self.environment_variable_provider = (
            environment_variable_provider or EnvironmentVariableProvider()
        )
        self.platform_helper_version_override = (
            platform_helper_version_override
            or self.environment_variable_provider.get(PLATFORM_HELPER_VERSION_OVERRIDE_KEY)
        )

    def generate(
        self,
        environment_name: str,
    ):

        config = self.config_provider.get_enriched_config()

        if environment_name not in config.get("environments").keys():
            raise EnvironmentNotFoundException(
                f"cannot generate terraform for environment {environment_name}.  It does not exist in your configuration"
            )

        platform_helper_version_for_template: str = config.get("default_versions", {}).get(
            "platform-helper"
        )

        if self.platform_helper_version_override:
            platform_helper_version_for_template = self.platform_helper_version_override

        module_source_override = self.environment_variable_provider.get(
            TERRAFORM_EXTENSIONS_MODULE_SOURCE_OVERRIDE_ENV_VAR
        )

        self.manifest_provider.generate_environment_config(
            config,
            environment_name,
            platform_helper_version_for_template,
            module_source_override,
        )



import os
from typing import Optional

from dbt_platform_helper.platform_exception import PlatformException


class EnvironmentVariableProvider:

    @staticmethod
    def get(env_var: str, default=None) -> Optional[str]:
        """Returns the stripped value or None if not set or empty."""
        value = os.environ.get(env_var)
        if value and value.strip():
            return value.strip()
        return default

    @staticmethod
    def get_required(env_var: str) -> str:
        """Returns the stripped value or raises a PlatformException if not set
        or empty."""
        value = os.environ.get(env_var)
        if not value or not value.strip():
            raise PlatformException(f"Environment variable '{env_var}' is not set or is empty")
        return value.strip()



import json
import stat
import subprocess
from collections.abc import Callable
from pathlib import Path
from typing import Tuple

import requests
import yaml
from boto3 import Session

from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.ecr import ECRProvider
from dbt_platform_helper.providers.files import FileProvider
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.parameter_store import ParameterStore
from dbt_platform_helper.utils.application import Application
from dbt_platform_helper.utils.application import (
    ApplicationEnvironmentNotFoundException,
)
from dbt_platform_helper.utils.application import load_application
from dbt_platform_helper.utils.aws import get_aws_session_or_abort
from dbt_platform_helper.utils.aws import get_build_url_from_arn
from dbt_platform_helper.utils.aws import get_build_url_from_pipeline_execution_id
from dbt_platform_helper.utils.aws import get_image_build_project
from dbt_platform_helper.utils.aws import get_manual_release_pipeline
from dbt_platform_helper.utils.aws import list_latest_images
from dbt_platform_helper.utils.aws import start_build_extraction
from dbt_platform_helper.utils.aws import start_pipeline_and_return_execution_id
from dbt_platform_helper.utils.template import setup_templates


class Codebase:
    def __init__(
        self,
        parameter_provider: ParameterStore,
        io: ClickIOProvider = ClickIOProvider(),
        load_application: Callable[[str], Application] = load_application,
        get_aws_session_or_abort: Callable[[str], Session] = get_aws_session_or_abort,
        ecr_provider: ECRProvider = ECRProvider(),
        get_image_build_project: Callable[[str], str] = get_image_build_project,
        get_manual_release_pipeline: Callable[[str], str] = get_manual_release_pipeline,
        get_build_url_from_arn: Callable[[str], str] = get_build_url_from_arn,
        get_build_url_from_pipeline_execution_id: Callable[
            [str], str
        ] = get_build_url_from_pipeline_execution_id,
        list_latest_images: Callable[[str], str] = list_latest_images,
        start_build_extraction: Callable[[str], str] = start_build_extraction,
        start_pipeline_and_return_execution_id: Callable[
            [str], str
        ] = start_pipeline_and_return_execution_id,
        run_subprocess: Callable[[str], str] = subprocess.run,
    ):
        self.parameter_provider = parameter_provider
        self.io = io
        self.load_application = load_application
        self.get_aws_session_or_abort = get_aws_session_or_abort
        self.ecr_provider = ecr_provider
        self.get_image_build_project = get_image_build_project
        self.get_manual_release_pipeline = get_manual_release_pipeline
        self.get_build_url_from_arn = get_build_url_from_arn
        self.get_build_url_from_pipeline_execution_id = get_build_url_from_pipeline_execution_id
        self.list_latest_images = list_latest_images
        self.start_build_extraction = start_build_extraction
        self.start_pipeline_and_return_execution_id = start_pipeline_and_return_execution_id
        self.run_subprocess = run_subprocess

    def prepare(self):
        """Sets up an application codebase for use within a DBT platform
        project."""
        templates = setup_templates()

        repository = (
            self.run_subprocess(
                ["git", "remote", "get-url", "origin"], capture_output=True, text=True
            )
            .stdout.split("/")[-1]
            .strip()
            .removesuffix(".git")
        )
        if repository.endswith("-deploy") or Path("./copilot").exists():
            raise NotInCodeBaseRepositoryException()

        builder_configuration_url = "https://raw.githubusercontent.com/uktrade/ci-image-builder/main/image_builder/configuration/builder_configuration.yml"
        builder_configuration_response = requests.get(builder_configuration_url)
        builder_configuration_content = yaml.safe_load(
            builder_configuration_response.content.decode("utf-8")
        )
        builder_versions = next(
            (
                item
                for item in builder_configuration_content["builders"]
                if item["name"] == "paketobuildpacks/builder-jammy-base"
            ),
            None,
        )
        builder_version = max(x["version"] for x in builder_versions["versions"])
        builder_version = min(builder_version, "0.4.240")

        Path("./.copilot/phases").mkdir(parents=True, exist_ok=True)
        image_build_run_contents = templates.get_template(f".copilot/image_build_run.sh").render()

        config_contents = templates.get_template(f".copilot/config.yml").render(
            repository=repository, builder_version=builder_version
        )
        self.io.info(
            FileProvider.mkfile(
                Path("."), ".copilot/image_build_run.sh", image_build_run_contents, overwrite=True
            )
        )

        image_build_run_file = Path(".copilot/image_build_run.sh")
        image_build_run_file.chmod(image_build_run_file.stat().st_mode | stat.S_IEXEC)

        self.io.info(
            FileProvider.mkfile(Path("."), ".copilot/config.yml", config_contents, overwrite=True)
        )

        for phase in ["build", "install", "post_build", "pre_build"]:
            phase_contents = templates.get_template(f".copilot/phases/{phase}.sh").render()

            self.io.info(
                FileProvider.mkfile(
                    Path("./.copilot"), f"phases/{phase}.sh", phase_contents, overwrite=True
                )
            )

    def build(self, app: str, codebase: str, commit: str):
        """Trigger a CodePipeline pipeline based build."""
        session = self.get_aws_session_or_abort()
        self.load_application(app, default_session=session)

        codebuild_client = session.client("codebuild")
        project_name = self.get_image_build_project(codebuild_client, app, codebase)
        build_url = self.__start_build_with_confirmation(
            codebuild_client,
            self.get_build_url_from_arn,
            f'You are about to build "{app}" for "{codebase}" with commit "{commit}". Do you want to continue?',
            {
                "projectName": project_name,
                "artifactsOverride": {"type": "NO_ARTIFACTS"},
                "sourceVersion": commit,
            },
        )

        if build_url:
            return self.io.info(
                f"Your build has been triggered. Check your build progress in the AWS Console: {build_url}"
            )

        raise ApplicationDeploymentNotTriggered(codebase)

    def deploy(
        self,
        app: str,
        env: str,
        codebase: str,
        commit: str = None,
        tag: str = None,
        branch: str = None,
    ):
        """Trigger a CodePipeline pipeline based deployment."""

        self._validate_reference_flags(commit, tag, branch)

        application, session = self._populate_application_values(app, env)

        image_ref = None
        if commit:
            self._validate_sha_length(commit)
            image_ref = f"commit-{commit}"
        elif tag:
            image_ref = f"tag-{tag}"
        elif branch:
            image_ref = f"branch-{branch}"

        image_ref = self.ecr_provider.get_commit_tag_for_reference(
            application.name, codebase, image_ref
        )

        codepipeline_client = session.client("codepipeline")
        pipeline_name = self.get_manual_release_pipeline(codepipeline_client, app, codebase)

        corresponding_to = ""
        if tag:
            corresponding_to = f"(corresponding to tag {tag}) "
        elif branch:
            corresponding_to = f"(corresponding to branch {branch}) "

        confirmation_message = f'\nYou are about to deploy "{app}" for "{codebase}" with image reference "{image_ref}" {corresponding_to}to the "{env}" environment using the "{pipeline_name}" deployment pipeline. Do you want to continue?'
        build_options = {
            "name": pipeline_name,
            "variables": [
                {"name": "ENVIRONMENT", "value": env},
                {"name": "IMAGE_TAG", "value": image_ref},
            ],
        }

        build_url = self.__start_pipeline_execution_with_confirmation(
            codepipeline_client,
            self.get_build_url_from_pipeline_execution_id,
            confirmation_message,
            build_options,
        )

        if build_url:
            return self.io.info(
                "Your deployment has been triggered. Check your build progress in the AWS Console: "
                f"{build_url}",
            )

        raise ApplicationDeploymentNotTriggered(codebase)

    def _validate_reference_flags(self, commit: str, tag: str, branch: str):
        provided = [ref for ref in [commit, tag, branch] if ref]

        if len(provided) == 0:
            self.io.abort_with_error(
                "To deploy, you must provide one of the options --commit, --tag or --branch."
            )
        elif len(provided) > 1:
            self.io.abort_with_error(
                "You have provided more than one of the --tag, --branch and --commit options but these are mutually exclusive. Please provide only one of these options."
            )

    def _populate_application_values(self, app: str, env: str) -> Tuple[Application, Session]:
        session = self.get_aws_session_or_abort()
        application = self.load_application(app, default_session=session)
        if not application.environments.get(env):
            raise ApplicationEnvironmentNotFoundException(application.name, env)
        return application, session

    def list(self, app: str, with_images: bool):
        """List available codebases for the application."""
        session = self.get_aws_session_or_abort()
        application = self.load_application(app, session)
        ecr_client = session.client("ecr")
        codebases = self.__get_codebases(application, session.client("ssm"))

        self.io.info("The following codebases are available:")

        for codebase in codebases:
            self.io.info(f"- {codebase['name']} (https://github.com/{codebase['repository']})")
            if with_images:
                self.list_latest_images(
                    ecr_client,
                    f"{application.name}/{codebase['name']}",
                    codebase["repository"],
                    self.io.info,
                )

        self.io.info("")

    def __get_codebases(self, application, ssm_client):
        parameters = self.parameter_provider.get_ssm_parameters_by_path(
            f"/copilot/applications/{application.name}/codebases"
        )
        codebases = [json.loads(p["Value"]) for p in parameters]

        if not codebases:
            return []
        return codebases

    def __start_build_with_confirmation(
        self,
        codebuild_client,
        get_build_url_from_arn,
        confirmation_message,
        build_options,
    ):
        if self.io.confirm(confirmation_message):
            build_arn = self.start_build_extraction(codebuild_client, build_options)
            return get_build_url_from_arn(build_arn)
        return None

    def __start_pipeline_execution_with_confirmation(
        self,
        codepipeline_client,
        get_build_url_from_pipeline_execution_id,
        confirmation_message,
        build_options,
    ):
        if self.io.confirm(confirmation_message):
            execution_id = self.start_pipeline_and_return_execution_id(
                codepipeline_client, build_options
            )
            return get_build_url_from_pipeline_execution_id(execution_id, build_options["name"])
        return None

    def _validate_sha_length(self, commit):
        if len(commit) < 7:
            self.io.abort_with_error(
                "Your commit reference is too short. Commit sha hashes specified by '--commit' must be at least 7 characters long."
            )


class ApplicationDeploymentNotTriggered(PlatformException):
    def __init__(self, codebase: str):
        super().__init__(f"""Your deployment for {codebase} was not triggered.""")


class NotInCodeBaseRepositoryException(PlatformException):
    def __init__(self):
        super().__init__(
            "You are in the deploy repository; make sure you are in the application codebase repository.",
        )



import itertools
import random
import re
import string
import traceback
from pathlib import Path
from typing import Callable
from typing import List
from typing import Union

import click

from dbt_platform_helper.constants import MAINTENANCE_PAGE_REASON
from dbt_platform_helper.constants import MANAGED_BY_PLATFORM
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.load_balancers import ListenerRuleNotFoundException
from dbt_platform_helper.providers.load_balancers import LoadBalancerProvider
from dbt_platform_helper.utils.application import Application
from dbt_platform_helper.utils.application import (
    ApplicationEnvironmentNotFoundException,
)
from dbt_platform_helper.utils.application import ApplicationServiceNotFoundException
from dbt_platform_helper.utils.application import Environment
from dbt_platform_helper.utils.application import Service


class MaintenancePageException(PlatformException):
    pass


class LoadBalancedWebServiceNotFoundException(MaintenancePageException):
    def __init__(self, application_name: str):
        super().__init__(f"No services deployed yet to {application_name} ")


class FailedToActivateMaintenancePageException(MaintenancePageException):
    def __init__(
        self,
        application_name: str,
        env: str,
        original_exception: Exception,
    ):
        super().__init__(
            f"Maintenance page failed to activate for the {application_name} application in environment {env}."
        )
        self.orginal_exception = original_exception

    def __str__(self):
        return f"{super().__str__()}\n" f"Original exception: {self.orginal_exception}"


# TODO: DBTP-1958: should this be in its own provider, inside the VPC one, what logic is this sepcific too?
def get_env_ips(vpc: str, application_environment: Environment) -> List[str]:
    account_name = f"{application_environment.session.profile_name}-vpc"
    vpc_name = vpc if vpc else account_name
    ssm_client = application_environment.session.client("ssm")

    try:
        param_value = ssm_client.get_parameter(Name=f"/{vpc_name}/EGRESS_IPS")["Parameter"]["Value"]
    except ssm_client.exceptions.ParameterNotFound:
        click.secho(f"No parameter found with name: /{vpc_name}/EGRESS_IPS")
        raise click.Abort

    return [ip.strip() for ip in param_value.split(",")]


class MaintenancePage:
    def __init__(
        self,
        application: Application,
        io: ClickIOProvider = ClickIOProvider(),
        load_balancer_provider: LoadBalancerProvider = LoadBalancerProvider,
        get_env_ips: Callable[[str, Environment], List[str]] = get_env_ips,
    ):
        self.application = application
        self.io = io
        self.load_balancer_provider = load_balancer_provider  # TODO: DBTP-1962: requires session from environment in application object which is only known during method execution
        self.load_balancer: LoadBalancerProvider = None
        self.get_env_ips = get_env_ips

    def _get_deployed_load_balanced_web_services(self, app: Application, svc: List[str]):
        if "*" in svc:
            services = [s for s in app.services.values() if s.kind == "Load Balanced Web Service"]
        else:
            all_services = [get_app_service(app, s) for s in list(svc)]
            services = [s for s in all_services if s.kind == "Load Balanced Web Service"]
        if not services:
            raise LoadBalancedWebServiceNotFoundException(app.name)
        return services

    # TODO: DBTP-1962: inject load balancer provider in activate method to avoid passing load balancer provider in init?
    def activate(self, env: str, services: List[str], template: str, vpc: Union[str, None]):

        services = self._get_deployed_load_balanced_web_services(self.application, services)
        application_environment = get_app_environment(self.application, env)
        self.load_balancer = self.load_balancer_provider(application_environment.session)

        https_listener = self.load_balancer.get_https_listener_for_application(
            self.application.name, env
        )
        current_maintenance_page = self.__get_maintenance_page_type(https_listener)
        remove_current_maintenance_page = False
        if current_maintenance_page:
            remove_current_maintenance_page = self.io.confirm(
                f"There is currently a '{current_maintenance_page}' maintenance page for the {env} "
                f"environment in {self.application.name}.\nWould you like to replace it with a '{template}' "
                f"maintenance page?"
            )
            if not remove_current_maintenance_page:
                return

        if remove_current_maintenance_page or self.io.confirm(
            f"You are about to enable the '{template}' maintenance page for the {env} "
            f"environment in {self.application.name}.\nWould you like to continue?"
        ):
            if current_maintenance_page and remove_current_maintenance_page:
                self.__remove_maintenance_page(https_listener)

            allowed_ips = self.get_env_ips(vpc, application_environment)

            self.add_maintenance_page(
                https_listener,
                self.application.name,
                env,
                services,
                allowed_ips,
                template,
            )
            self.io.info(
                f"Maintenance page '{template}' added for environment {env} in application {self.application.name}",
            )

    def deactivate(self, env: str):
        application_environment = get_app_environment(self.application, env)

        self.load_balancer = self.load_balancer_provider(application_environment.session)

        https_listener = self.load_balancer.get_https_listener_for_application(
            self.application.name, env
        )
        current_maintenance_page = self.__get_maintenance_page_type(https_listener)

        if not current_maintenance_page:
            self.io.warn("There is no current maintenance page to remove")
            return

        if not self.io.confirm(
            f"There is currently a '{current_maintenance_page}' maintenance page, "
            f"would you like to remove it?"
        ):
            return

        self.__remove_maintenance_page(https_listener)
        self.io.info(
            f"Maintenance page removed from environment {env} in application {self.application.name}",
        )

    def add_maintenance_page(
        self,
        listener_arn: str,
        app: str,
        env: str,
        services: List[Service],
        allowed_ips: List[str],
        template: str = "default",
    ):

        maintenance_page_content = get_maintenance_page_template(template)
        bypass_value = "".join(random.choices(string.ascii_lowercase + string.digits, k=12))

        rule_priority = itertools.count(start=1)
        maintenance_page_host_header_conditions = []
        try:
            for svc in services:
                target_group_arn = self.load_balancer.find_target_group(app, env, svc.name)

                # not all of an application's services are guaranteed to have been deployed to an environment
                if not target_group_arn:
                    continue

                service_conditions = self.load_balancer.get_host_header_conditions(
                    listener_arn, target_group_arn
                )

                self.io.debug(
                    f"""
#----------------------------------------------------------#
# Creating listener rules for service {svc.name.ljust(21, " ")}#
#----------------------------------------------------------#

""",
                )

                for ip in allowed_ips:
                    self.load_balancer.create_header_rule(
                        listener_arn,
                        target_group_arn,
                        "X-Forwarded-For",
                        [ip],
                        "AllowedIps",
                        next(rule_priority),
                        service_conditions,
                        [
                            {"Key": "application", "Value": app},
                            {"Key": "environment", "Value": env},
                            {"Key": "reason", "Value": MAINTENANCE_PAGE_REASON},
                            {"Key": "managed-by", "Value": MANAGED_BY_PLATFORM},
                            {"Key": "service", "Value": svc.name},
                        ],
                    )
                    self.load_balancer.create_source_ip_rule(
                        listener_arn,
                        target_group_arn,
                        [ip],
                        "AllowedSourceIps",
                        next(rule_priority),
                        service_conditions,
                        [
                            {"Key": "application", "Value": app},
                            {"Key": "environment", "Value": env},
                            {"Key": "reason", "Value": MAINTENANCE_PAGE_REASON},
                            {"Key": "managed-by", "Value": MANAGED_BY_PLATFORM},
                            {"Key": "service", "Value": svc.name},
                        ],
                    )

                self.load_balancer.create_header_rule(
                    listener_arn,
                    target_group_arn,
                    "Bypass-Key",
                    [bypass_value],
                    "BypassIpFilter",
                    next(rule_priority),
                    service_conditions,
                    [
                        {"Key": "application", "Value": app},
                        {"Key": "environment", "Value": env},
                        {"Key": "reason", "Value": MAINTENANCE_PAGE_REASON},
                        {"Key": "managed-by", "Value": MANAGED_BY_PLATFORM},
                        {"Key": "service", "Value": svc.name},
                    ],
                )

                # add to accumilating list of conditions for maintenace page rule
                maintenance_page_host_header_conditions.extend(service_conditions)

            self.io.info(
                f"\nUse a browser plugin to add `Bypass-Key` header with value {bypass_value} to your requests. For more detail, visit https://platform.readme.trade.gov.uk/next-steps/put-a-service-under-maintenance/",
            )

            unique_sorted_host_headers = sorted(
                list(
                    {
                        value
                        for condition in maintenance_page_host_header_conditions
                        for value in condition["HostHeaderConfig"]["Values"]
                    }
                )
            )

            # Can only set 4 host headers per rule as listener rules have a max conditions of 5
            for i in range(0, len(unique_sorted_host_headers), 4):
                self.load_balancer.create_rule(
                    listener_arn=listener_arn,
                    priority=next(rule_priority),
                    conditions=[
                        {
                            "Field": "path-pattern",
                            "PathPatternConfig": {"Values": ["/*"]},
                        },
                        {
                            "Field": "host-header",
                            "HostHeaderConfig": {
                                "Values": unique_sorted_host_headers[i : i + 4],
                            },
                        },
                    ],
                    actions=[
                        {
                            "Type": "fixed-response",
                            "FixedResponseConfig": {
                                "StatusCode": "503",
                                "ContentType": "text/html",
                                "MessageBody": maintenance_page_content,
                            },
                        }
                    ],
                    tags=[
                        {"Key": "application", "Value": app},
                        {"Key": "environment", "Value": env},
                        {"Key": "reason", "Value": MAINTENANCE_PAGE_REASON},
                        {"Key": "name", "Value": "MaintenancePage"},
                        {"Key": "type", "Value": template},
                        {"Key": "managed-by", "Value": MANAGED_BY_PLATFORM},
                    ],
                )
        except Exception as e:
            self.__clean_up_maintenance_page_rules(listener_arn)
            raise FailedToActivateMaintenancePageException(
                app, env, f"{e}:\n {traceback.format_exc()}"
            )

    def __clean_up_maintenance_page_rules(
        self, listener_arn: str, fail_when_not_deleted: bool = False
    ) -> None:

        tag_descriptions = self.load_balancer.get_rules_tag_descriptions_by_listener_arn(
            listener_arn
        )

        # keep track of rules deleted
        deleted_rules = {"MaintenancePage": 0}
        for name in ["MaintenancePage", "AllowedIps", "BypassIpFilter", "AllowedSourceIps"]:
            deleted_list = self.load_balancer.delete_listener_rule_by_tags(tag_descriptions, name)

            # track the rules deleted grouped by service
            for deleted_rule in deleted_list:
                tags = {t["Key"]: t["Value"] for t in deleted_rule["Tags"]}
                if "service" in tags:
                    if tags["service"] not in deleted_rules:
                        deleted_rules[tags["service"]] = {
                            "AllowedIps": 0,
                            "BypassIpFilter": 0,
                            "AllowedSourceIps": 0,
                        }
                    deleted_rules[tags["service"]][name] += 1
                elif tags.get("name") == "MaintenancePage":
                    deleted_rules["MaintenancePage"] += 1

            if (
                fail_when_not_deleted
                and name == "MaintenancePage"
                and deleted_rules["MaintenancePage"] == 0
            ):
                raise ListenerRuleNotFoundException()

        self.io.warn(
            f"Rules deleted by type and grouped by service: {deleted_rules}",
        )

    def __remove_maintenance_page(self, listener_arn: str) -> dict[str, bool]:
        self.__clean_up_maintenance_page_rules(listener_arn, True)

    def __get_maintenance_page_type(self, listener_arn: str) -> Union[str, None]:
        tag_descriptions = self.load_balancer.get_rules_tag_descriptions_by_listener_arn(
            listener_arn
        )
        maintenance_page_type = None
        for description in tag_descriptions:
            tags = {t["Key"]: t["Value"] for t in description["Tags"]}
            if tags.get("name") == "MaintenancePage":
                maintenance_page_type = tags.get("type")

        return maintenance_page_type


def get_app_service(application: Application, svc_name: str) -> Service:
    application_service = application.services.get(svc_name)

    if not application_service:
        raise ApplicationServiceNotFoundException(application.name, svc_name)

    return application_service


def get_app_environment(application: Application, env_name: str) -> Environment:
    application_environment = application.environments.get(env_name)

    if not application_environment:
        raise ApplicationEnvironmentNotFoundException(application.name, env_name)

    return application_environment


def get_maintenance_page_template(template) -> str:
    template_contents = (
        Path(__file__)
        .parent.parent.joinpath(
            f"templates/svc/maintenance_pages/{template}.html",
        )
        .read_text()
        .replace("\n", "")
    )

    # [^\S]\s+ - Remove any space that is not preceded by a non-space character.
    return re.sub(r"[^\S]\s+", "", template_contents)



import json
import subprocess
import time

from botocore.exceptions import ClientError

from dbt_platform_helper.constants import CONDUIT_DOCKER_IMAGE_LOCATION
from dbt_platform_helper.providers.aws.exceptions import CreateTaskTimeoutException
from dbt_platform_helper.providers.secrets import Secrets
from dbt_platform_helper.utils.application import Application
from dbt_platform_helper.utils.messages import abort_with_error


def create_addon_client_task(
    iam_client,
    ssm_client,
    application: Application,
    env: str,
    addon_type: str,
    addon_name: str,
    task_name: str,
    access: str,
):
    secret_name = f"/copilot/{application.name}/{env}/secrets/{_normalise_secret_name(addon_name)}"

    if addon_type == "postgres":
        if access == "read":
            secret_name += "_READ_ONLY_USER"
        elif access == "write":
            secret_name += "_APPLICATION_USER"
        elif access == "admin":
            create_postgres_admin_task(
                ssm_client,
                application,
                addon_name,
                addon_type,
                env,
                secret_name,
                task_name,
            )
            return
    elif addon_type == "redis" or addon_type == "opensearch":
        secret_name += "_ENDPOINT"

    role_name = f"{addon_name}-{application.name}-{env}-conduitEcsTask"

    try:
        iam_client.get_role(RoleName=role_name)
        execution_role = f"--execution-role {role_name} "
    except ClientError as ex:
        execution_role = ""
        # We cannot check for botocore.errorfactory.NoSuchEntityException as botocore generates that class on the fly as part of errorfactory.
        # factory. Checking the error code is the recommended way of handling these exceptions.
        if ex.response.get("Error", {}).get("Code", None) != "NoSuchEntity":
            # TODO: DBTP-1946: When we are refactoring this, raise an exception to be caught at the command layer
            abort_with_error(
                f"cannot obtain Role {role_name}: {ex.response.get('Error', {}).get('Message', '')}"
            )

    subprocess.call(
        f"copilot task run --app {application.name} --env {env} "
        f"--task-group-name {task_name} "
        f"{execution_role}"
        f"--image {CONDUIT_DOCKER_IMAGE_LOCATION}:{addon_type} "
        f"--secrets CONNECTION_SECRET={_get_secrets_provider(application, env).get_connection_secret_arn(secret_name)} "
        "--cpu 2048 --memory 4096 "
        "--platform-os linux "
        "--platform-arch arm64",
        shell=True,
    )


def get_postgres_admin_connection_string(
    ssm_client, secret_name: str, app: Application, env: str, addon_name: str
):
    read_only_secret_name = secret_name + "_READ_ONLY_USER"
    master_secret_name = (
        f"/copilot/{app.name}/{env}/secrets/{_normalise_secret_name(addon_name)}_RDS_MASTER_ARN"
    )
    master_secret_arn = ssm_client.get_parameter(Name=master_secret_name, WithDecryption=True)[
        "Parameter"
    ]["Value"]
    connection_string = json.dumps(
        _get_secrets_provider(app, env).get_postgres_connection_data_updated_with_master_secret(
            read_only_secret_name, master_secret_arn
        )
    )

    return connection_string


def create_postgres_admin_task(
    ssm_client,
    app: Application,
    addon_name: str,
    addon_type: str,
    env: str,
    secret_name: str,
    task_name: str,
):

    connection_string = get_postgres_admin_connection_string(
        ssm_client, secret_name, app, env, addon_name
    )

    subprocess.call(
        f"copilot task run --app {app.name} --env {env} "
        f"--task-group-name {task_name} "
        f"--image {CONDUIT_DOCKER_IMAGE_LOCATION}:{addon_type} "
        f"--env-vars CONNECTION_SECRET='{connection_string}' "
        "--cpu 2048 --memory 4096 "
        "--platform-os linux "
        "--platform-arch arm64",
        shell=True,
    )


def _temp_until_refactor_get_ecs_task_arns(ecs_client, cluster_arn: str, task_name: str):
    tasks = ecs_client.list_tasks(
        cluster=cluster_arn,
        desiredStatus="RUNNING",
        family=f"copilot-{task_name}",
    )

    if not tasks["taskArns"]:
        return []

    return tasks["taskArns"]


def connect_to_addon_client_task(
    ecs_client,
    application_name,
    env,
    cluster_arn,
    task_name,
    get_ecs_task_arns=_temp_until_refactor_get_ecs_task_arns,
):
    running = False
    tries = 0
    while tries < 15 and not running:
        tries += 1
        # TODO: DBTP-1946: Use from ECS provider when we refactor this
        if get_ecs_task_arns(ecs_client, cluster_arn, task_name):
            subprocess.call(
                "copilot task exec "
                f"--app {application_name} --env {env} "
                f"--name {task_name} "
                f"--command bash",
                shell=True,
            )
            running = True

        time.sleep(1)

    if not running:
        raise CreateTaskTimeoutException(task_name, application_name, env)


def _normalise_secret_name(addon_name: str) -> str:
    return addon_name.replace("-", "_").upper()


def _get_secrets_provider(application: Application, env: str) -> Secrets:
    # TODO: DBTP-1946: We instantiate the secrets provider here to avoid rabbit holing, but something better probably possible when we are refactoring this area
    return Secrets(
        application.environments[env].session.client("ssm"),
        application.environments[env].session.client("secretsmanager"),
        application.name,
        env,
    )



import json
import os
import time
import urllib.parse
from collections import OrderedDict
from copy import deepcopy
from datetime import datetime
from datetime import timezone
from importlib.metadata import version
from pathlib import Path
from typing import Any

from dbt_platform_helper.constants import IMAGE_TAG_ENV_VAR
from dbt_platform_helper.constants import PLATFORM_HELPER_PACKAGE_NAME
from dbt_platform_helper.constants import PLATFORM_HELPER_VERSION_OVERRIDE_KEY
from dbt_platform_helper.constants import SERVICE_CONFIG_FILE
from dbt_platform_helper.constants import SERVICE_DIRECTORY
from dbt_platform_helper.constants import (
    TERRAFORM_ECS_SERVICE_MODULE_SOURCE_OVERRIDE_ENV_VAR,
)
from dbt_platform_helper.constants import TERRAFORM_MODULE_SOURCE_TYPE_ENV_VAR
from dbt_platform_helper.domain.terraform_environment import (
    EnvironmentNotFoundException,
)
from dbt_platform_helper.entities.service import ServiceConfig
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.autoscaling import AutoscalingProvider
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.config_validator import ConfigValidator
from dbt_platform_helper.providers.ecs import ECS
from dbt_platform_helper.providers.environment_variable import (
    EnvironmentVariableProvider,
)
from dbt_platform_helper.providers.files import FileProvider
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.logs import LogsProvider
from dbt_platform_helper.providers.s3 import S3Provider
from dbt_platform_helper.providers.terraform_manifest import TerraformManifestProvider
from dbt_platform_helper.providers.version import InstalledVersionProvider
from dbt_platform_helper.providers.yaml_file import YamlFileProvider
from dbt_platform_helper.utils.application import load_application
from dbt_platform_helper.utils.deep_merge import deep_merge

SERVICE_TYPES = ["Load Balanced Web Service", "Backend Service"]
DEPLOYMENT_TIMEOUT_SECONDS = 1200
POLL_INTERVAL_SECONDS = 5

# TODO add schema version to service config


class ServiceManager:
    def __init__(
        self,
        config_provider=ConfigProvider(ConfigValidator()),
        io: ClickIOProvider = ClickIOProvider(),
        file_provider=YamlFileProvider,
        manifest_provider: TerraformManifestProvider = None,
        platform_helper_version_override: str = None,
        load_application=load_application,
        installed_version_provider: InstalledVersionProvider = InstalledVersionProvider(),
        ecs_provider: ECS = None,
        s3_provider: S3Provider = None,
        logs_provider: LogsProvider = None,
        autoscaling_provider: AutoscalingProvider = None,
    ):

        self.file_provider = file_provider
        self.config_provider = config_provider
        self.io = io
        self.manifest_provider = manifest_provider or TerraformManifestProvider()
        self.platform_helper_version_override = (
            platform_helper_version_override
            or EnvironmentVariableProvider.get(PLATFORM_HELPER_VERSION_OVERRIDE_KEY)
        )
        self.load_application = load_application
        self.installed_version_provider = installed_version_provider
        self.ecs_provider = ecs_provider
        self.s3_provider = s3_provider
        self.logs_provider = logs_provider
        self.autoscaling_provider = autoscaling_provider

    def generate(self, environment: str, services: list[str]):

        config = self.config_provider.get_enriched_config()
        application_name = config.get("application", "")
        application = self.load_application(app=application_name)

        if environment not in application.environments:
            raise EnvironmentNotFoundException(
                f"Cannot generate Terraform for environment '{environment}'. It does not exist in your configuration."
            )

        if not services:
            try:
                for dir in Path("services").iterdir():
                    if dir.is_dir():
                        config_path = dir / SERVICE_CONFIG_FILE
                        if config_path.exists():
                            services.append(dir.name)
                        else:
                            self.io.warn(
                                f"Failed loading service name from {dir.name}.\n"
                                "Please ensure that your '/services' directory follows the correct structure (i.e. /services/<service_name>/service-config.yml) and the 'service-config.yml' contents are correct."
                            )
            except Exception as e:
                self.io.abort_with_error(f"Failed extracting services with exception, {e}")

        service_models = []
        for service in services:
            file_content = self.file_provider.load(
                f"{SERVICE_DIRECTORY}/{service}/{SERVICE_CONFIG_FILE}"
            )

            file_content = YamlFileProvider.find_and_replace(
                config=file_content,
                strings=[
                    "${PLATFORM_APPLICATION_NAME}",
                    "${PLATFORM_ENVIRONMENT_NAME}",
                ],
                replacements=[application.name, environment],
            )
            service_models.append(ServiceConfig(**file_content))

        platform_helper_version_for_template: str = (
            self.platform_helper_version_override
            or config.get("default_versions", {}).get("platform-helper")
        )

        source_type = EnvironmentVariableProvider.get(TERRAFORM_MODULE_SOURCE_TYPE_ENV_VAR)

        if source_type == "LOCAL":
            module_source_override = ServiceConfig.local_terraform_source
        elif source_type == "OVERRIDE":
            module_source_override = EnvironmentVariableProvider.get(
                TERRAFORM_ECS_SERVICE_MODULE_SOURCE_OVERRIDE_ENV_VAR
            )
        else:
            module_source_override = None

        timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        for service in service_models:

            model_dump = service.model_dump(
                exclude_none=True,
                by_alias=True,
                mode="json",
            )  # Use by_alias=True so that the Cooldown field 'in_' is written as 'in' in the output
            env_overrides = model_dump.get("environments", {}).get(environment)
            if env_overrides:
                merged_config = deep_merge(model_dump, env_overrides)
            else:
                merged_config = model_dump.copy()
            merged_config.pop("environments", None)

            output_path = Path(
                f"terraform/{SERVICE_DIRECTORY}/{environment}/{service.name}/{SERVICE_CONFIG_FILE}"
            )
            output_path.parent.mkdir(parents=True, exist_ok=True)

            self.file_provider.write(
                str(output_path),
                merged_config,
                f"# WARNING: This is an autogenerated file, not for manual editing.\n# Generated by platform-helper {version('dbt-platform-helper')} / {timestamp}.\n",
            )

            self.manifest_provider.generate_service_config(
                service,
                environment,
                platform_helper_version_for_template,
                config,
                module_source_override,
            )

    def migrate_copilot_manifests(self) -> None:
        service_directory = Path("services/")
        service_directory.mkdir(parents=True, exist_ok=True)

        for dirname, _, filenames in os.walk("copilot"):
            if "manifest.yml" in filenames and "environments" not in dirname:
                copilot_manifest = self.file_provider.load(f"{dirname}/manifest.yml")
                service_manifest = OrderedDict(deepcopy(copilot_manifest))

                if service_manifest["type"] not in SERVICE_TYPES:
                    continue

                if "environments" in service_manifest:
                    for env in service_manifest["environments"]:
                        env_config = service_manifest["environments"][env]
                        if "http" in env_config:
                            if "alb" in env_config["http"]:
                                del env_config["http"]["alb"]

                if "entrypoint" in service_manifest:
                    if isinstance(service_manifest["entrypoint"], str):
                        service_manifest["entrypoint"] = [service_manifest["entrypoint"]]

                service_manifest = self.file_provider.find_and_replace(
                    config=service_manifest,
                    strings=["${COPILOT_APPLICATION_NAME}", "${COPILOT_ENVIRONMENT_NAME}"],
                    replacements=["${PLATFORM_APPLICATION_NAME}", "${PLATFORM_ENVIRONMENT_NAME}"],
                )

                service_manifest = self.file_provider.remove_empty_keys(service_manifest)

                if "sidecars" in service_manifest:
                    new_sidecars = {}
                    writable_directories = []

                    for sidecar_name, sidecar in service_manifest["sidecars"].items():
                        if "chown" not in sidecar.get("command", "") and "chmod" not in sidecar.get(
                            "command", ""
                        ):
                            new_sidecars[sidecar_name] = sidecar
                        if "chown" in sidecar.get("command", "") and "mount_points" in sidecar:
                            for mountpoint in sidecar["mount_points"]:
                                writable_directories.append(mountpoint["path"])

                    service_manifest["sidecars"] = new_sidecars
                    if "storage" in service_manifest:
                        service_manifest["storage"]["writable_directories"] = writable_directories

                service_path = service_directory / service_manifest["name"]

                self.io.info(
                    FileProvider.mkfile(
                        service_path,
                        "service-config.yml",
                        "",
                        overwrite=True,
                    )
                )

                current_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                message = f"# Generated by platform-helper {self.installed_version_provider.get_semantic_version(PLATFORM_HELPER_PACKAGE_NAME)} / {current_date}.\n\n"

                self.file_provider.write(
                    f"{service_path}/service-config.yml", dict(service_manifest), message
                )

    def deploy(
        self,
        service: str,
        environment: str,
        application: str,
        image_tag: str = None,
    ):
        """Register a new ECS task definition revision, update the ECS service
        with it, monitor service, task and container logs, and wait until
        deployment is complete."""

        start_time = datetime.now(timezone.utc)
        cluster_name = f"{application}-{environment}-cluster"
        ecs_service_name = f"{application}-{environment}-{service}"

        s3_response = self.s3_provider.get_object(
            bucket_name=f"ecs-task-definitions-{application}-{environment}",
            object_key=f"{application}/{environment}/{service}.json",
        )

        task_definition = json.loads(s3_response)

        image_tag = image_tag or EnvironmentVariableProvider.get(IMAGE_TAG_ENV_VAR)

        self.io.info(
            f"Deploying image tag '{image_tag}' to service '{ecs_service_name}' in environment '{environment}'.\n"
        )

        task_def_arn = self.ecs_provider.register_task_definition(
            service=service,
            image_tag=image_tag,
            task_definition=task_definition,
        )

        self._output_with_timestamp(
            f"Task definition successfully registered with ARN '{task_def_arn}'."
        )

        autoscaling_response = self.autoscaling_provider.describe_autoscaling_target(
            cluster_name=cluster_name, ecs_service_name=ecs_service_name
        )
        desired_count = autoscaling_response.get("MinCapacity", 1)

        update_response = self.ecs_provider.update_service(
            service=service,
            task_def_arn=task_def_arn,
            environment=environment,
            application=application,
            desired_count=desired_count,
        )

        self._output_with_timestamp(
            f"Successfully updated ECS service '{update_response['serviceName']}'."
        )

        primary_deployment_id = self._get_primary_deployment_id(service_response=update_response)

        self._output_with_timestamp(
            f"New deployment with ID '{primary_deployment_id}' has been triggered."
        )

        seen_events = set()
        deadline = time.monotonic() + DEPLOYMENT_TIMEOUT_SECONDS

        while time.monotonic() < deadline:
            service_response = self.ecs_provider.describe_service(
                application=application, environment=environment, service=service
            )
            primary_deployment_id = self._get_primary_deployment_id(
                service_response=service_response
            )

            self._monitor_service_events(
                service_response=service_response, seen_events=seen_events, start_time=start_time
            )

            task_ids = self._wait_for_new_tasks(
                cluster_name=cluster_name, deployment_id=primary_deployment_id
            )
            task_response = self.ecs_provider.describe_tasks(
                cluster_name=f"{application}-{environment}-cluster",
                task_ids=task_ids,
            )

            log_group = f"/platform/ecs/service/{application}/{environment}/{service}"
            self._monitor_task_events(
                task_response=task_response, seen_events=seen_events, log_group=log_group
            )

            state, reason = self.ecs_provider.get_service_deployment_state(
                cluster_name=cluster_name,
                service_name=ecs_service_name,
                start_time=start_time.timestamp(),
            )

            if state == "SUCCESSFUL":
                self._output_with_timestamp("Deployment complete.")
                return
            if state in ["STOPPED", "ROLLBACK_SUCCESSFUL", "ROLLBACK_FAILED"]:
                raise PlatformException(f"Deployment failed: {reason or 'unknown reason'}")

            time.sleep(POLL_INTERVAL_SECONDS)

        raise PlatformException("Timed out waiting for service to stabilise.")

    @staticmethod
    def _get_primary_deployment_id(service_response: dict[str, Any]):
        for dep in service_response["deployments"]:
            if dep["status"] == "PRIMARY":
                return dep["id"]
        raise PlatformException(
            f"Unable to find primary ECS deployment for service '{service_response['serviceName']}'."
        )

    def _monitor_service_events(
        self, service_response: dict[str, Any], seen_events: set[str], start_time: datetime
    ):
        """Output ECS service events during deployment."""

        for event in reversed(service_response.get("events", [])):
            if event["id"] not in seen_events and event["createdAt"] > start_time:
                seen_events.add(event["id"])
                timestamp = event["createdAt"].strftime("%H:%M:%S")
                message = event["message"]
                self._output_with_timestamp(
                    message=message,
                    error=("error" in message or "failed" in message),
                    timestamp=timestamp,
                )

    def _monitor_task_events(
        self, task_response: list[dict[str, Any]], seen_events: set[str], log_group: str
    ):
        """Output ECS task and container errors during deployment."""

        for task in task_response:
            for container in task["containers"]:
                if container.get("exitCode", 0) != 1:
                    continue

                task_id = task["taskArn"].split("/")[-1]
                container_name = container["name"]
                log_stream = f"platform/{container_name}/{task_id}"

                if f"{task_id}-{container_name}" not in seen_events:
                    seen_events.add(f"{task_id}-{container_name}")
                    self._output_with_timestamp(
                        message=f"Container '{container_name}' stopped in task '{task_id}'.",
                        error=True,
                    )
                    log_url = f"https://eu-west-2.console.aws.amazon.com/cloudwatch/home?region=eu-west-2#logsV2:log-groups/log-group/{urllib.parse.quote_plus(log_group)}/log-events/{urllib.parse.quote_plus(log_stream)}"
                    self._output_with_timestamp(
                        message=f"View CloudWatch log: {log_url}", error=True
                    )

                log_events = self.logs_provider.get_log_stream_events(
                    log_group=log_group,
                    log_stream=log_stream,
                    limit=20,
                )

                for event in log_events:
                    try:
                        message = json.loads(event["message"])
                    except json.decoder.JSONDecodeError:
                        message = event["message"]

                    if f"{task_id}-{message}" not in seen_events:
                        seen_events.add(f"{task_id}-{message}")
                        self._output_with_timestamp(message=message, error=True)

    def _wait_for_new_tasks(self, cluster_name: str, deployment_id: str) -> list[str]:
        """Return first ECS task ID started by the PRIMARY ECS deployment."""

        timeout_seconds = 180
        deadline = time.monotonic() + timeout_seconds

        while time.monotonic() < deadline:
            task_arns = self.ecs_provider.get_ecs_task_arns(
                cluster=cluster_name,
                started_by=deployment_id,
                desired_status="RUNNING",
            )

            if task_arns:
                break

            time.sleep(POLL_INTERVAL_SECONDS)

        if not task_arns:
            raise PlatformException(
                f"Timed out waiting for RUNNING ECS tasks to spin up after {timeout_seconds}s."
            )

        task_ids = []
        for arn in task_arns:
            task_ids.append(arn.rsplit("/", 1)[-1])
        return task_ids

    def _output_with_timestamp(self, message: str, error: bool = False, timestamp: datetime = None):
        if not timestamp:
            timestamp = datetime.now(timezone.utc).strftime("%H:%M:%S")

        if error:
            self.io.deploy_error(f"[{timestamp}] {message}")
        else:
            self.io.info(f"[{timestamp}] {message}")



from copy import deepcopy
from datetime import datetime
from pathlib import Path

from schema import SchemaError

from dbt_platform_helper.constants import FIRST_UPGRADABLE_PLATFORM_HELPER_MAJOR_VERSION
from dbt_platform_helper.constants import PLATFORM_CONFIG_FILE
from dbt_platform_helper.constants import PLATFORM_CONFIG_SCHEMA_VERSION
from dbt_platform_helper.constants import PLATFORM_HELPER_PACKAGE_NAME
from dbt_platform_helper.entities.platform_config_schema import PlatformConfigSchema
from dbt_platform_helper.entities.semantic_version import SemanticVersion
from dbt_platform_helper.providers.config_validator import ConfigValidator
from dbt_platform_helper.providers.config_validator import ConfigValidatorError
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.version import InstalledVersionProvider
from dbt_platform_helper.providers.yaml_file import FileNotFoundException
from dbt_platform_helper.providers.yaml_file import FileProviderException
from dbt_platform_helper.providers.yaml_file import YamlFileProvider

SCHEMA_VERSION_MESSAGE = """Installed version: platform-helper: {installed_platform_helper_version} (schema version: {installed_schema_version})
'platform-config.yml' version: platform-helper: {config_platform_helper_version} (schema version: {config_schema_version})"""
PLEASE_UPGRADE_TO_V13_MESSAGE = """Please ensure that you have already upgraded to platform-helper 13, following the instructions in https://platform.readme.trade.gov.uk/reference/upgrading-platform-helper/.

Then upgrade platform-helper to version {installed_platform_helper_version} and run 'platform-helper config migrate' to upgrade the configuration to the current schema version."""


class ConfigLoader:
    def __init__(self, file_provider=YamlFileProvider, io: ClickIOProvider = ClickIOProvider()):
        self.io = io
        self.file_provider = file_provider

    def load(self, path):
        try:
            file_content = self.file_provider.load(path)
            return file_content
        except FileNotFoundException as e:
            self.io.abort_with_error(
                f"{e} Please check it exists and you are in the root directory of your -deploy repository."
            )
        except FileProviderException as e:
            self.io.abort_with_error(f"Error loading configuration from {path}: {e}")


class ConfigProvider:
    def __init__(
        self,
        config_validator: ConfigValidator = ConfigValidator(),
        file_provider: YamlFileProvider = YamlFileProvider,
        io: ClickIOProvider = ClickIOProvider(),
        schema_version_for_installed_platform_helper: int = PLATFORM_CONFIG_SCHEMA_VERSION,
        installed_version_provider: InstalledVersionProvider = InstalledVersionProvider,
    ):
        self.config = {}
        self.validator = config_validator
        self.io = io
        self.file_provider = file_provider
        self.schema_version_for_installed_platform_helper = (
            schema_version_for_installed_platform_helper
        )
        self.installed_version_provider = installed_version_provider

    # TODO: DBTP-1964: refactor so that apply_environment_defaults isn't set, discarded and set again
    def get_enriched_config(self):
        return self.apply_environment_defaults(self.load_and_validate_platform_config())

    def _validate_platform_config(self):
        PlatformConfigSchema.schema().validate(self.config)
        # TODO: DBTP-1964: = logically this isn't validation but loading + parsing, to move.
        # also, we apply defaults but discard that data.  Should we just apply
        # defaults to config returned by load_and_validate
        enriched_config = ConfigProvider.apply_environment_defaults(self.config)

        try:
            self.validator.run_validations(enriched_config)
        except ConfigValidatorError as exc:
            self.io.abort_with_error(f"Config validation has failed.\n{str(exc)}")

    def load_and_validate_platform_config(self, path=PLATFORM_CONFIG_FILE):
        try:
            self.config = self.file_provider.load(path)
        except FileNotFoundException as e:
            self.io.abort_with_error(
                f"{e} Please check it exists and you are in the root directory of your deployment project."
            )
        except FileProviderException as e:
            self.io.abort_with_error(f"Error loading configuration from {path}: {e}")

        self._validate_schema_version()

        try:
            self._validate_platform_config()
        except SchemaError as e:
            self.io.abort_with_error(f"Schema error in {path}. {e}")

        return self.config

    def _abort_due_to_schema_version_error(self, config_description: str, action_required: str):
        self.io.abort_with_error(
            "\n".join(
                [
                    config_description,
                    "",
                    action_required,
                ]
            )
        )

    def _validate_schema_version(self):
        config_schema_version = self.config.get("schema_version")
        config_platform_helper_version = self.config.get("default_versions", {}).get(
            "platform-helper", ""
        )
        header = SCHEMA_VERSION_MESSAGE.format(
            installed_platform_helper_version=self._installed_platform_helper_version(),
            installed_schema_version=self.schema_version_for_installed_platform_helper,
            config_platform_helper_version=(
                config_platform_helper_version if config_platform_helper_version else "N/A"
            ),
            config_schema_version=(config_schema_version if config_schema_version else "N/A"),
        )

        if config_schema_version:
            self._handle_schema_version_mismatch(config_schema_version, header)
        else:
            self._handle_missing_schema_version(config_platform_helper_version, header)

    def _handle_schema_version_mismatch(self, platform_config_schema_version: int, header: str):
        platform_config_schema_version_is_old = (
            platform_config_schema_version < self.schema_version_for_installed_platform_helper
        )
        installed_platform_helper_is_old = (
            platform_config_schema_version > self.schema_version_for_installed_platform_helper
        )

        if platform_config_schema_version_is_old:
            self._abort_due_to_schema_version_error(
                header,
                "Please upgrade your platform-config.yml by running 'platform-helper config migrate'.",
            )
        elif installed_platform_helper_is_old:
            self._abort_due_to_schema_version_error(
                header,
                f"Please update your platform-helper to a version that supports schema_version: {platform_config_schema_version}.",
            )
        # else the schema_version is the correct one so continue.

    def _handle_missing_schema_version(self, config_platform_helper_version: str, header: str):
        config_p_h_version_semver = SemanticVersion.from_string(config_platform_helper_version)
        major_version = config_p_h_version_semver and config_p_h_version_semver.major
        platform_config_is_old_but_supported_by_migrations = (
            major_version and major_version == FIRST_UPGRADABLE_PLATFORM_HELPER_MAJOR_VERSION
        )
        platform_config_is_old = (
            major_version and major_version < FIRST_UPGRADABLE_PLATFORM_HELPER_MAJOR_VERSION
        )
        platform_config_is_really_old = not major_version
        installed_platform_helper_version = self._installed_platform_helper_version()

        if platform_config_is_old_but_supported_by_migrations:
            self._abort_due_to_schema_version_error(
                header,
                f"Please upgrade your platform-config.yml to be compatible with {installed_platform_helper_version} by running: 'platform-helper config migrate'.",
            )
        elif platform_config_is_old or platform_config_is_really_old:
            self._abort_due_to_schema_version_error(
                header,
                PLEASE_UPGRADE_TO_V13_MESSAGE.format(
                    installed_platform_helper_version=installed_platform_helper_version,
                ),
            )
        # if major_version and major_version > FIRST_UPGRADABLE_PLATFORM_HELPER_MAJOR_VERSION then
        # the platform-config.yml is malformed and so should progress to validation if appropriate.

    def load_unvalidated_config_file(self, path=PLATFORM_CONFIG_FILE):
        try:
            return self.file_provider.load(path)
        except FileProviderException:
            return {}

    # TODO: DBTP-1888: remove function and push logic to where this is called.
    # removed usage from config domain, code is very generic and doesn't require the overhead of a function
    def config_file_check(self, path=PLATFORM_CONFIG_FILE):
        if not Path(path).exists():
            self.io.abort_with_error(
                f"`{path}` is missing. "
                "Please check it exists and you are in the root directory of your deployment project."
            )

    @staticmethod
    def apply_environment_defaults(config):
        if "environments" not in config:
            return config

        enriched_config = deepcopy(config)

        environments = enriched_config["environments"]
        env_defaults = environments.get("*", {})
        without_defaults_entry = {
            name: data if data else {} for name, data in environments.items() if name != "*"
        }

        config.get("default_versions", {})

        def combine_env_data(data):
            return {
                **env_defaults,
                **data,
            }

        defaulted_envs = {
            env_name: combine_env_data(env_data)
            for env_name, env_data in without_defaults_entry.items()
        }

        enriched_config["environments"] = defaulted_envs

        return enriched_config

    def write_platform_config(self, new_platform_config):
        current_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        message = f"# Generated by platform-helper {self._installed_platform_helper_version()} / {current_date}.\n\n"
        self.file_provider.write(PLATFORM_CONFIG_FILE, new_platform_config, message)

    def _installed_platform_helper_version(self) -> str:
        return str(
            self.installed_version_provider.get_semantic_version(PLATFORM_HELPER_PACKAGE_NAME)
        )



import os
import re
import webbrowser
from pathlib import Path
from typing import Dict

from prettytable import PrettyTable

from dbt_platform_helper.constants import PLATFORM_CONFIG_FILE
from dbt_platform_helper.constants import STANDARD_PLATFORM_SSO_ROLES
from dbt_platform_helper.domain.versioning import AWSVersioning
from dbt_platform_helper.domain.versioning import CopilotVersioning
from dbt_platform_helper.domain.versioning import PlatformHelperVersioning
from dbt_platform_helper.entities.semantic_version import (
    IncompatibleMajorVersionException,
)
from dbt_platform_helper.entities.semantic_version import SemanticVersion
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.aws.sso_auth import SSOAuthProvider
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.schema_migrator import ALL_MIGRATIONS
from dbt_platform_helper.providers.schema_migrator import Migrator
from dbt_platform_helper.providers.validation import ValidationException
from dbt_platform_helper.providers.version_status import VersionStatus

yes = "\033[92m✔\033[0m"
no = "\033[91m✖\033[0m"
maybe = "\033[93m?\033[0m"

RECOMMENDATIONS = {
    "dbt-platform-helper-upgrade": (
        "Upgrade dbt-platform-helper to version {version} `pip install "
        "--upgrade dbt-platform-helper=={version}`."
    ),
    "dbt-platform-helper-upgrade-note": (
        "Post upgrade, run `platform-helper copilot make-addons` to " "update your addon templates."
    ),
    "generic-tool-upgrade": "Upgrade {tool} to version {version}.",
    "install-copilot": "Install AWS Copilot https://aws.github.io/copilot-cli/",
    "install-aws": "Install AWS CLI https://aws.amazon.com/cli/",
}

AWS_CONFIG = """
#
# uktrade
#

[sso-session uktrade]
sso_start_url = https://uktrade.awsapps.com/start#/
sso_region = eu-west-2
sso_registration_scopes = sso:account:access

[default]
sso_session = uktrade
region = eu-west-2
output = json

"""


class NoDeploymentRepoConfigException(PlatformException):
    def __init__(self):
        super().__init__("Could not find a deployment repository, no checks to run.")


# TODO: DBTP-1993: move to generic location so it can be reused
class NoPlatformConfigException(PlatformException):
    def __init__(self):
        super().__init__(
            f"`platform-config.yml` is missing. "
            "Please check it exists and you are in the root directory of your deployment project."
        )


class Config:
    def __init__(
        self,
        io: ClickIOProvider = ClickIOProvider(),
        platform_helper_versioning: PlatformHelperVersioning = PlatformHelperVersioning(),
        aws_versioning: AWSVersioning = AWSVersioning(),
        copilot_versioning: CopilotVersioning = CopilotVersioning(),
        sso: SSOAuthProvider = None,
        config_provider: ConfigProvider = ConfigProvider(),
        migrator: Migrator = Migrator(ALL_MIGRATIONS, io_provider=ClickIOProvider()),
    ):
        self.oidc_app = None
        self.io = io
        self.platform_helper_versioning = platform_helper_versioning
        self.aws_versioning = aws_versioning
        self.copilot_versioning = copilot_versioning
        self.sso = sso or SSOAuthProvider()
        self.SSO_START_URL = "https://uktrade.awsapps.com/start"
        self.config_provider = config_provider
        self.migrator = migrator

    def validate(self):
        if not Path("copilot").exists():
            raise NoDeploymentRepoConfigException()
        if not Path(PLATFORM_CONFIG_FILE).exists():
            raise NoPlatformConfigException()

        self.io.debug("\nDetected a deployment repository\n")
        platform_helper_version_status = self.platform_helper_versioning.get_version_status()
        aws_version_status = self.aws_versioning.get_version_status()
        copilot_version_status = self.copilot_versioning.get_version_status()

        self._check_tool_versions(
            platform_helper_version_status, aws_version_status, copilot_version_status
        )

        compatible = self._check_addon_versions(platform_helper_version_status)

        exit(0 if compatible else 1)

    def migrate(self):
        platform_config = self.config_provider.load_unvalidated_config_file()
        new_platform_config = self.migrator.migrate(platform_config)
        self.config_provider.write_platform_config(new_platform_config)

    def generate_aws(self, file_path):
        self.oidc_app = self._create_oidc_application()
        verification_url, device_code = self._get_device_code(self.oidc_app)

        if self.io.confirm(
            "You are about to be redirected to a verification page. You will need to complete sign-in before returning to the command line. Do you want to continue?",
        ):
            webbrowser.open(verification_url)

        if self.io.confirm(
            "Have you completed the sign-in process in your browser?",
        ):
            access_token = self.sso.create_access_token(
                client_id=self.oidc_app[0],
                client_secret=self.oidc_app[1],
                device_code=device_code,
            )

        aws_config_path = os.path.expanduser(file_path)

        if self.io.confirm(
            f"This command is destructive and will overwrite file contents at {file_path}. Are you sure you want to continue?"
        ):
            self.io.info(
                "Fetching credentials... this may take longer if you have access to many accounts."
            )

            with open(aws_config_path, "w") as config_file:
                config_file.write(AWS_CONFIG)

                for account in self._retrieve_aws_accounts(access_token):
                    config_file.write(f"[profile {account['accountName']}]\n")
                    config_file.write("sso_session = uktrade\n")
                    config_file.write(f"sso_account_id = {account['accountId']}\n")
                    config_file.write(
                        f"sso_role_name = {self._retrieve_role_for_aws_account(aws_sso_token=access_token, account_id=account['accountId'])}\n"
                    )
                    config_file.write("region = eu-west-2\n")
                    config_file.write("output = json\n")
                    config_file.write("\n")

    def _create_oidc_application(self):
        self.io.debug("Creating temporary AWS SSO OIDC application")
        client_id, client_secret = self.sso.register(
            client_name="platform-helper",
            client_type="public",
        )
        return client_id, client_secret

    def _get_device_code(self, oidc_application):
        self.io.debug("Initiating device code flow")
        url, device_code = self.sso.start_device_authorization(
            client_id=oidc_application[0],
            client_secret=oidc_application[1],
            start_url=self.SSO_START_URL,
        )

        return url, device_code

    def _retrieve_aws_accounts(self, aws_sso_token):
        accounts_list = self.sso.list_accounts(
            access_token=aws_sso_token,
            max_results=100,
        )
        return accounts_list

    def _retrieve_role_for_aws_account(self, aws_sso_token, account_id):
        """
        Selects the most appropriate IAM role for a given AWS account.

        Roles listed in STANDARD_PLATFORM_SSO_ROLES are preferred if available.
        If none are found, the first available role returned by
        sso.list_account_roles is used.
        """

        role_list = self.sso.list_account_roles(
            access_token=aws_sso_token,
            account_id=account_id,
            max_results=100,
        )

        roles_retrieved = [r["roleName"] for r in role_list]

        for role in STANDARD_PLATFORM_SSO_ROLES:
            if role in roles_retrieved:
                return role
        return role_list[0]["roleName"]

    def _add_version_status_row(
        self, table: PrettyTable, header: str, version_status: VersionStatus
    ):
        table.add_row(
            [
                header,
                str(version_status.installed),
                str(version_status.latest),
                no if version_status.is_outdated() else yes,
            ]
        )

    def _check_tool_versions(
        self,
        platform_helper_version_status: VersionStatus,
        aws_version_status: VersionStatus,
        copilot_version_status: VersionStatus,
    ):
        self.io.debug("Checking tooling versions...")

        recommendations = {}

        if copilot_version_status.installed is None:
            recommendations["install-copilot"] = RECOMMENDATIONS["install-copilot"]

        if aws_version_status.installed is None:
            recommendations["install-aws"] = RECOMMENDATIONS["install-aws"]

        tool_versions_table = PrettyTable()
        tool_versions_table.field_names = [
            "Tool",
            "Local version",
            "Released version",
            "Running latest?",
        ]
        tool_versions_table.align["Tool"] = "l"

        self._add_version_status_row(tool_versions_table, "aws", aws_version_status)
        self._add_version_status_row(tool_versions_table, "copilot", copilot_version_status)
        self._add_version_status_row(
            tool_versions_table, "dbt-platform-helper", platform_helper_version_status
        )

        self.io.info(tool_versions_table)

        if aws_version_status.is_outdated() and "install-aws" not in recommendations:
            recommendations["aws-upgrade"] = RECOMMENDATIONS["generic-tool-upgrade"].format(
                tool="AWS CLI",
                version=str(aws_version_status.latest),
            )

        if copilot_version_status.is_outdated() and "install-copilot" not in recommendations:
            recommendations["copilot-upgrade"] = RECOMMENDATIONS["generic-tool-upgrade"].format(
                tool="AWS Copilot",
                version=str(copilot_version_status.latest),
            )

        if platform_helper_version_status.is_outdated():
            recommendations["dbt-platform-helper-upgrade"] = RECOMMENDATIONS[
                "dbt-platform-helper-upgrade"
            ].format(version=str(platform_helper_version_status.latest))
            recommendations["dbt-platform-helper-upgrade-note"] = RECOMMENDATIONS[
                "dbt-platform-helper-upgrade-note"
            ]

        self._render_recommendations(recommendations)

    def _check_addon_versions(self, platform_helper_versions: VersionStatus) -> bool:

        self.io.debug("Checking addons templates versions...")

        compatible = True
        recommendations = {}

        local_version = platform_helper_versions.installed
        latest_release = platform_helper_versions.latest

        addons_templates_table = PrettyTable()
        addons_templates_table.field_names = [
            "Addons Template File",
            "Generated with",
            "Compatible with local?",
            "Compatible with latest?",
        ]
        addons_templates_table.align["Addons Template File"] = "l"

        addons_templates = list(Path("./copilot").glob("**/addons/*"))
        # Sort by template file path
        addons_templates.sort(key=lambda e: str(e))
        # Bring environment addons to the top
        addons_templates.sort(key=lambda e: "environments/" not in str(e))

        for template_file in addons_templates:
            generated_with_version = maybe
            local_compatible_symbol = yes
            latest_compatible_symbol = yes

            generated_with_version = None

            try:
                generated_with_version = self.__get_template_generated_with_version(
                    str(template_file.resolve())
                )
            except ValidationException:
                local_compatible_symbol = maybe
                compatible = False
                recommendations["dbt-platform-helper-upgrade"] = RECOMMENDATIONS[
                    "dbt-platform-helper-upgrade"
                ].format(version=latest_release)
                recommendations["dbt-platform-helper-upgrade-note"] = RECOMMENDATIONS[
                    "dbt-platform-helper-upgrade-note"
                ]

            try:
                local_version.validate_compatibility_with(generated_with_version)
            except IncompatibleMajorVersionException:
                local_compatible_symbol = no
                compatible = False
                recommendations["dbt-platform-helper-upgrade"] = RECOMMENDATIONS[
                    "dbt-platform-helper-upgrade"
                ].format(version=latest_release)
                recommendations["dbt-platform-helper-upgrade-note"] = RECOMMENDATIONS[
                    "dbt-platform-helper-upgrade-note"
                ]
            except ValidationException:
                local_compatible_symbol = maybe
                compatible = False

            try:
                latest_release.validate_compatibility_with(generated_with_version)
            except IncompatibleMajorVersionException:
                latest_compatible_symbol = no
                compatible = False
            except ValidationException:
                latest_compatible_symbol = maybe
                compatible = False

            addons_templates_table.add_row(
                [
                    template_file.relative_to("."),
                    (maybe if latest_compatible_symbol is maybe else str(generated_with_version)),
                    local_compatible_symbol,
                    latest_compatible_symbol,
                ]
            )

        self.io.info(addons_templates_table)
        self._render_recommendations(recommendations)

        return compatible

    def _render_recommendations(self, recommendations: Dict[str, str]):
        if recommendations:
            self.io.info("\nRecommendations:\n", bold=True)

            for name, recommendation in recommendations.items():
                if name.endswith("-note"):
                    continue
                self.io.info(f"  - {recommendation}")
                if recommendations.get(f"{name}-note", False):
                    self.io.info(f"    {recommendations.get(f'{name}-note')}")

            self.io.info("")

    def __get_template_generated_with_version(self, template_file_path: str) -> SemanticVersion:
        try:
            template_contents = Path(template_file_path).read_text()
            template_version = re.search(
                r"# Generated by platform-helper ([v.\-0-9]+)", template_contents
            ).group(1)
            return SemanticVersion.from_string(template_version)
        except (IndexError, AttributeError):
            raise ValidationException(f"Template {template_file_path} has no version information")



import click

from dbt_platform_helper.commands.environment import AVAILABLE_TEMPLATES
from dbt_platform_helper.domain.database_copy import DatabaseCopy
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.utils.click import ClickDocOptGroup


@click.group(chain=True, cls=ClickDocOptGroup)
def database():
    """Commands to copy data between databases."""


@database.command(name="dump")
@click.option(
    "--app",
    type=str,
    help="The application name. Required unless you are running the command from your deploy repo",
)
@click.option(
    "--from",
    "from_env",
    type=str,
    required=True,
    help="The environment you are dumping data from",
)
@click.option(
    "--database", type=str, required=True, help="The name of the database you are dumping data from"
)
@click.option(
    "--from-vpc",
    type=str,
    help="The vpc the specified environment is running in. Required unless you are running the command from your deploy repo",
)
@click.option(
    "--filename",
    type=str,
    help="Specify a name for the database dump file. Recommended if the same dump database is being used for multiple load environments",
)
def dump(app, from_env, database, from_vpc, filename):
    """Dump a database into an S3 bucket."""
    try:
        data_copy = DatabaseCopy(app, database)
        data_copy.dump(from_env, from_vpc, filename)
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))


@database.command(name="load")
@click.option(
    "--app",
    type=str,
    help="The application name. Required unless you are running the command from your deploy repo",
)
@click.option(
    "--to", "to_env", type=str, required=True, help="The environment you are loading data into"
)
@click.option(
    "--database", type=str, required=True, help="The name of the database you are loading data into"
)
@click.option(
    "--to-vpc",
    type=str,
    help="The vpc the specified environment is running in. Required unless you are running the command from your deploy repo",
)
@click.option("--auto-approve/--no-auto-approve", default=False)
@click.option(
    "--filename",
    type=str,
    help="Specify a name for the database dump file. Recommended if the same dump database is being used for multiple load environments",
)
def load(app, to_env, database, to_vpc, auto_approve, filename):
    """Load a database from an S3 bucket."""
    try:
        data_copy = DatabaseCopy(app, database, auto_approve)
        data_copy.load(to_env, to_vpc, filename)
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))


@database.command(name="copy")
@click.option(
    "--app",
    type=str,
    help="The application name. Required unless you are running the command from your deploy repo",
)
@click.option(
    "--from", "from_env", type=str, required=True, help="The environment you are copying data from"
)
@click.option(
    "--to", "to_env", type=str, required=True, help="The environment you are copying data into"
)
@click.option(
    "--database", type=str, required=True, help="The name of the database you are copying"
)
@click.option(
    "--from-vpc",
    type=str,
    help="The vpc the environment you are copying from is running in. Required unless you are running the command from your deploy repo",
)
@click.option(
    "--to-vpc",
    type=str,
    help="The vpc the environment you are copying into is running in. Required unless you are running the command from your deploy repo",
)
@click.option("--auto-approve/--no-auto-approve", default=False)
@click.option("--svc", type=str, required=True, multiple=True, default=["web"])
@click.option(
    "--template",
    type=click.Choice(AVAILABLE_TEMPLATES),
    default="default",
    help="The maintenance page you wish to put up.",
)
@click.option("--no-maintenance-page", flag_value=True)
def copy(
    app,
    from_env,
    to_env,
    database,
    from_vpc,
    to_vpc,
    auto_approve,
    svc,
    template,
    no_maintenance_page,
):
    """Copy a database between environments."""
    try:
        data_copy = DatabaseCopy(app, database, auto_approve)
        data_copy.copy(from_env, to_env, from_vpc, to_vpc, svc, template, no_maintenance_page)
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))



import click

from dbt_platform_helper.domain.copilot_environment import CopilotEnvironment
from dbt_platform_helper.domain.maintenance_page import MaintenancePage
from dbt_platform_helper.domain.terraform_environment import TerraformEnvironment
from dbt_platform_helper.domain.versioning import PlatformHelperVersioning
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.cloudformation import CloudFormation
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.config_validator import ConfigValidator
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.vpc import VpcProvider
from dbt_platform_helper.utils.application import load_application
from dbt_platform_helper.utils.aws import get_aws_session_or_abort
from dbt_platform_helper.utils.click import ClickDocOptGroup

AVAILABLE_TEMPLATES = ["default", "migration", "dmas-migration"]


@click.group(cls=ClickDocOptGroup)
def environment():
    """Commands affecting environments."""
    PlatformHelperVersioning().check_if_needs_update()


@environment.command()
@click.option("--app", type=str, required=True)
@click.option("--env", type=str, required=True)
@click.option("--svc", type=str, required=True, multiple=True, default=["web"])
@click.option(
    "--template",
    type=click.Choice(AVAILABLE_TEMPLATES),
    default="default",
    help="The maintenance page you wish to put up.",
)
@click.option("--vpc", type=str)
def offline(app, env, svc, template, vpc):
    """Take load-balanced web services offline with a maintenance page."""

    try:
        application = load_application(app)
        MaintenancePage(application).activate(env, svc, template, vpc)
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))


@environment.command()
@click.option("--app", type=str, required=True)
@click.option("--env", type=str, required=True)
def online(app, env):
    """Remove a maintenance page from an environment."""

    try:
        application = load_application(app)
        MaintenancePage(application).deactivate(env)
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))


@environment.command()
@click.option(
    "--name",
    "-n",
    required=True,
    help="The name of the environment to generate a copilot manifest for.",
)
def generate(name):
    """Gathers various IDs and ARNs from AWS and generates the AWS Copilot
    environment manifest at copilot/environments/<environment>/manifest.yml."""

    click_io = ClickIOProvider()
    try:
        session = get_aws_session_or_abort()
        config_provider = ConfigProvider(ConfigValidator())
        vpc_provider = VpcProvider(session)
        cloudformation_provider = CloudFormation(session.client("cloudformation"))

        CopilotEnvironment(
            config_provider, vpc_provider, cloudformation_provider, session
        ).generate(name)
    except PlatformException as err:
        click_io.abort_with_error(str(err))


@environment.command(help="Generate terraform manifest for the specified environment.")
@click.option(
    "--name", "-n", required=True, help="The name of the environment to generate a manifest for."
)
def generate_terraform(name):
    click_io = ClickIOProvider()
    try:
        session = get_aws_session_or_abort()
        config_provider = ConfigProvider(ConfigValidator(session=session))
        TerraformEnvironment(config_provider).generate(name)

    except PlatformException as err:
        click_io.abort_with_error(str(err))
